__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging
import mimetypes
import os
import re
import sys
from urllib.parse import urlparse

import tornado
import tornado.escape
import tornado.gen
import tornado.http1connection
import tornado.httpclient
import tornado.httpserver
import tornado.httputil
import tornado.iostream
import tornado.tcpserver
import tornado.util
import tornado.web
from zipstream.ng import ZIP_DEFLATED, ZipStream

import octoprint.util


def fix_json_encode():
    """
    This makes tornado.escape.json_encode use octoprint.util.JsonEncoding.encode as fallback in order to allow
    serialization of globally registered types like frozendict and others.
    """

    import json

    from octoprint.util.json import JsonEncoding

    def fixed_json_encode(value):
        return json.dumps(value, default=JsonEncoding.encode, allow_nan=False).replace(
            "</", "<\\/"
        )

    import tornado.escape

    tornado.escape.json_encode = fixed_json_encode


def enable_per_message_deflate_extension():
    """
    This configures tornado.websocket.WebSocketHandler.get_compression_options to support the permessage-deflate extension
    to the websocket protocol, minimizing data bandwidth if clients support the extension as well
    """

    def get_compression_options(self):
        return {"compression_level": 1, "mem_level": 1}

    tornado.websocket.WebSocketHandler.get_compression_options = get_compression_options


def fix_websocket_check_origin():
    """
    This fixes tornado.websocket.WebSocketHandler.check_origin to do the same origin check against the Host
    header case-insensitively, as defined in RFC6454, Section 4, item 5.
    """

    scheme_translation = {"wss": "https", "ws": "http"}

    def patched_check_origin(self, origin):
        def get_check_tuple(urlstring):
            parsed = urlparse(urlstring)
            scheme = scheme_translation.get(parsed.scheme, parsed.scheme)
            return (
                scheme,
                parsed.hostname,
                parsed.port
                if parsed.port
                else 80
                if scheme == "http"
                else 443
                if scheme == "https"
                else None,
            )

        return get_check_tuple(origin) == get_check_tuple(self.request.full_url())

    import tornado.websocket

    tornado.websocket.WebSocketHandler.check_origin = patched_check_origin


# ~~ More sensible logging


class RequestlessExceptionLoggingMixin(tornado.web.RequestHandler):

    LOG_REQUEST = False

    def log_exception(self, typ, value, tb, *args, **kwargs):
        if isinstance(value, tornado.web.HTTPError):
            if value.log_message:
                format = "%d %s: " + value.log_message
                args = [value.status_code, self._request_summary()] + list(value.args)
                tornado.web.gen_log.warning(format, *args)
        else:
            if self.LOG_REQUEST:
                tornado.web.app_log.error(
                    "Uncaught exception %s\n%r",
                    self._request_summary(),
                    self.request,
                    exc_info=(typ, value, tb),
                )
            else:
                tornado.web.app_log.error(
                    "Uncaught exception %s",
                    self._request_summary(),
                    exc_info=(typ, value, tb),
                )


# ~~ CORS support


class CorsSupportMixin(tornado.web.RequestHandler):
    """
    `tornado.web.RequestHandler <http://tornado.readthedocs.org/en/branch4.0/web.html#request-handlers>`_ mixin that
    makes sure to set CORS headers similarly to the Flask backed API endpoints.
    """

    ENABLE_CORS = False

    def set_default_headers(self):
        origin = self.request.headers.get("Origin")
        if self.request.method != "OPTIONS" and origin and self.ENABLE_CORS:
            self.set_header("Access-Control-Allow-Origin", origin)

    @tornado.gen.coroutine
    def options(self, *args, **kwargs):
        if self.ENABLE_CORS:
            origin = self.request.headers.get("Origin")
            method = self.request.headers.get("Access-Control-Request-Method")

            # Allow the origin which made the XHR
            self.set_header("Access-Control-Allow-Origin", origin)
            # Allow the actual method
            self.set_header("Access-Control-Allow-Methods", method)
            # Allow for 10 seconds
            self.set_header("Access-Control-Max-Age", "10")

            # 'preflight' request contains the non-standard headers the real request will have (like X-Api-Key)
            custom_headers = self.request.headers.get("Access-Control-Request-Headers")
            if custom_headers is not None:
                self.set_header("Access-Control-Allow-Headers", custom_headers)

        self.set_status(204)
        self.finish()


# ~~ WSGI middleware


@tornado.web.stream_request_body
class UploadStorageFallbackHandler(RequestlessExceptionLoggingMixin, CorsSupportMixin):
    """
    A ``RequestHandler`` similar to ``tornado.web.FallbackHandler`` which fetches any files contained in the request bodies
    of content type ``multipart``, stores them in temporary files and supplies the ``fallback`` with the file's ``name``,
    ``content_type``, ``path`` and ``size`` instead via a rewritten body.

    Basically similar to what the nginx upload module does.

    Basic request body example:

    .. code-block:: none

        ------WebKitFormBoundarypYiSUx63abAmhT5C
        Content-Disposition: form-data; name="file"; filename="test.gcode"
        Content-Type: application/octet-stream

        ...
        ------WebKitFormBoundarypYiSUx63abAmhT5C
        Content-Disposition: form-data; name="apikey"

        my_funny_apikey
        ------WebKitFormBoundarypYiSUx63abAmhT5C
        Content-Disposition: form-data; name="select"

        true
        ------WebKitFormBoundarypYiSUx63abAmhT5C--

    That would get turned into:

    .. code-block:: none

        ------WebKitFormBoundarypYiSUx63abAmhT5C
        Content-Disposition: form-data; name="apikey"

        my_funny_apikey
        ------WebKitFormBoundarypYiSUx63abAmhT5C
        Content-Disposition: form-data; name="select"

        true
        ------WebKitFormBoundarypYiSUx63abAmhT5C
        Content-Disposition: form-data; name="file.path"
        Content-Type: text/plain; charset=utf-8

        /tmp/tmpzupkro
        ------WebKitFormBoundarypYiSUx63abAmhT5C
        Content-Disposition: form-data; name="file.name"
        Content-Type: text/plain; charset=utf-8

        test.gcode
        ------WebKitFormBoundarypYiSUx63abAmhT5C
        Content-Disposition: form-data; name="file.content_type"
        Content-Type: text/plain; charset=utf-8

        application/octet-stream
        ------WebKitFormBoundarypYiSUx63abAmhT5C
        Content-Disposition: form-data; name="file.size"
        Content-Type: text/plain; charset=utf-8

        349182
        ------WebKitFormBoundarypYiSUx63abAmhT5C--

    The underlying application can then access the contained files via their respective paths and just move them
    where necessary.
    """

    BODY_METHODS = ("POST", "PATCH", "PUT")
    """ The request methods that may contain a request body. """

    def initialize(
        self, fallback, file_prefix="tmp", file_suffix="", path=None, suffixes=None
    ):
        if not suffixes:
            suffixes = {}

        self._fallback = fallback
        self._file_prefix = file_prefix
        self._file_suffix = file_suffix
        self._path = path

        self._suffixes = {key: key for key in ("name", "path", "content_type", "size")}
        for suffix_type, suffix in suffixes.items():
            if suffix_type in self._suffixes and suffix is not None:
                self._suffixes[suffix_type] = suffix

        # multipart boundary
        self._multipart_boundary = None

        # Parts, files and values will be stored here
        self._parts = {}
        self._files = []

        # Part currently being processed
        self._current_part = None

        # content type of request body
        self._content_type = None

        # bytes left to read according to content_length of request body
        self._bytes_left = 0

        # buffer needed for identifying form data parts
        self._buffer = b""

        # buffer for new body
        self._new_body = b""

        # logger
        self._logger = logging.getLogger(__name__)

    def prepare(self):
        """
        Prepares the processing of the request. If it's a request that may contain a request body (as defined in
        :attr:`UploadStorageFallbackHandler.BODY_METHODS`) prepares the multipart parsing if content type fits. If it's a
        body-less request, just calls the ``fallback`` with an empty body and finishes the request.
        """
        if self.request.method in UploadStorageFallbackHandler.BODY_METHODS:
            self._bytes_left = self.request.headers.get("Content-Length", 0)
            self._content_type = self.request.headers.get("Content-Type", None)

            # request might contain a body
            if self.is_multipart():
                if not self._bytes_left:
                    # we don't support requests without a content-length
                    raise tornado.web.HTTPError(
                        411, log_message="No Content-Length supplied"
                    )

                # extract the multipart boundary
                fields = self._content_type.split(";")
                for field in fields:
                    k, sep, v = field.strip().partition("=")
                    if k == "boundary" and v:
                        if v.startswith('"') and v.endswith('"'):
                            self._multipart_boundary = tornado.escape.utf8(v[1:-1])
                        else:
                            self._multipart_boundary = tornado.escape.utf8(v)
                        break
                else:
                    # RFC2046 section 5.1 (as referred to from RFC 7578) defines the boundary
                    # parameter as mandatory for multipart requests:
                    #
                    #     The only mandatory global parameter for the "multipart" media type is
                    #     the boundary parameter, which consists of 1 to 70 characters [...]
                    #
                    # So no boundary? 400 Bad Request
                    raise tornado.web.HTTPError(
                        400, log_message="No multipart boundary supplied"
                    )
        else:
            self._fallback(self.request, b"")
            self._finished = True

    def data_received(self, chunk):
        """
        Called by Tornado on receiving a chunk of the request body. If request is a multipart request, takes care of
        processing the multipart data structure via :func:`_process_multipart_data`. If not, just adds the chunk to
        internal in-memory buffer.

        :param chunk: chunk of data received from Tornado
        """

        data = self._buffer + chunk
        if self.is_multipart():
            self._process_multipart_data(data)
        else:
            self._buffer = data

    def is_multipart(self):
        """Checks whether this request is a ``multipart`` request"""
        return self._content_type is not None and self._content_type.startswith(
            "multipart"
        )

    def _process_multipart_data(self, data):
        """
        Processes the given data, parsing it for multipart definitions and calling the appropriate methods.

        :param data: the data to process as a string
        """

        # check for boundary
        delimiter = b"--%s" % self._multipart_boundary
        delimiter_loc = data.find(delimiter)
        delimiter_len = len(delimiter)
        end_of_header = -1
        if delimiter_loc != -1:
            # found the delimiter in the currently available data
            delimiter_data_end = 0 if delimiter_loc == 0 else delimiter_loc - 2
            data, self._buffer = data[0:delimiter_data_end], data[delimiter_loc:]
            end_of_header = self._buffer.find(b"\r\n\r\n")
        else:
            # make sure any boundary (with single or double ==) contained at the end of chunk does not get
            # truncated by this processing round => save it to the buffer for next round
            endlen = len(self._multipart_boundary) + 4
            data, self._buffer = data[0:-endlen], data[-endlen:]

        # stream data to part handler
        if data and self._current_part:
            self._on_part_data(self._current_part, data)

        if end_of_header >= 0:
            self._on_part_header(self._buffer[delimiter_len + 2 : end_of_header])
            self._buffer = self._buffer[end_of_header + 4 :]

        if delimiter_loc != -1 and self._buffer.strip() == delimiter + b"--":
            # we saw the last boundary and are at the end of our request
            if self._current_part:
                self._on_part_finish(self._current_part)
                self._current_part = None
            self._buffer = b""
            self._on_request_body_finish()

    def _on_part_header(self, header):
        """
        Called for a new multipart header, takes care of parsing the header and calling :func:`_on_part` with the
        relevant data, setting the current part in the process.

        :param header: header to parse
        """

        # close any open parts
        if self._current_part:
            self._on_part_finish(self._current_part)
            self._current_part = None

        header_check = header.find(self._multipart_boundary)
        if header_check != -1:
            self._logger.warning(
                "Header still contained multipart boundary, stripping it..."
            )
            header = header[header_check:]

        # convert to dict
        try:
            header = tornado.httputil.HTTPHeaders.parse(header.decode("utf-8"))
        except UnicodeDecodeError:
            try:
                header = tornado.httputil.HTTPHeaders.parse(header.decode("iso-8859-1"))
            except Exception:
                # looks like we couldn't decode something here neither as UTF-8 nor ISO-8859-1
                self._logger.warning(
                    "Could not decode multipart headers in request, should be either UTF-8 or ISO-8859-1"
                )
                self.send_error(400)
                return

        disp_header = header.get("Content-Disposition", "")
        disposition, disp_params = _parse_header(disp_header, strip_quotes=False)

        if disposition != "form-data":
            self._logger.warning(
                "Got a multipart header without form-data content disposition, ignoring that one"
            )
            return
        if not disp_params.get("name"):
            self._logger.warning("Got a multipart header without name, ignoring that one")
            return

        filename = disp_params.get("filename*", None)  # RFC 5987 header present?
        if filename is not None:
            try:
                filename = _extended_header_value(filename)
            except Exception:
                # parse error, this is not RFC 5987 compliant after all
                self._logger.warning(
                    "extended filename* value {!r} is not RFC 5987 compliant".format(
                        filename
                    )
                )
                self.send_error(400)
                return
        else:
            # no filename* header, just strip quotes from filename header then and be done
            filename = _strip_value_quotes(disp_params.get("filename", None))

        self._current_part = self._on_part_start(
            _strip_value_quotes(disp_params["name"]),
            header.get("Content-Type", None),
            filename=filename,
        )

    def _on_part_start(self, name, content_type, filename=None):
        """
        Called for new parts in the multipart stream. If ``filename`` is given creates new ``file`` part (which leads
        to storage of the data as temporary file on disk), if not creates a new ``data`` part (which stores
        incoming data in memory).

        Structure of ``file`` parts:

        * ``name``: name of the part
        * ``filename``: filename associated with the part
        * ``path``: path to the temporary file storing the file's data
        * ``content_type``: content type of the part
        * ``file``: file handle for the temporary file (mode "wb", not deleted on close, will be deleted however after
          handling of the request has finished in :func:`_handle_method`)

        Structure of ``data`` parts:

        * ``name``: name of the part
        * ``content_type``: content type of the part
        * ``data``: bytes of the part (initialized to an empty string)

        :param name: name of the part
        :param content_type: content type of the part
        :param filename: filename associated with the part.
        :return: dict describing the new part
        """
        if filename is not None:
            # this is a file
            import tempfile

            handle = tempfile.NamedTemporaryFile(
                mode="wb",
                prefix=self._file_prefix,
                suffix=self._file_suffix,
                dir=self._path,
                delete=False,
            )
            return {
                "name": tornado.escape.utf8(name),
                "filename": tornado.escape.utf8(filename),
                "path": tornado.escape.utf8(handle.name),
                "content_type": tornado.escape.utf8(content_type),
                "file": handle,
            }

        else:
            return {
                "name": tornado.escape.utf8(name),
                "content_type": tornado.escape.utf8(content_type),
                "data": b"",
            }

    def _on_part_data(self, part, data):
        """
        Called when new bytes are received for the given ``part``, takes care of writing them to their storage.

        :param part: part for which data was received
        :param data: data chunk which was received
        """
        if "file" in part:
            part["file"].write(data)
        else:
            part["data"] += data

    def _on_part_finish(self, part):
        """
        Called when a part gets closed, takes care of storing the finished part in the internal parts storage and for
        ``file`` parts closing the temporary file and storing the part in the internal files storage.

        :param part: part which was closed
        """
        name = part["name"]
        self._parts[name] = part
        if "file" in part:
            self._files.append(part["path"])
            part["file"].close()
            del part["file"]

    def _on_request_body_finish(self):
        """
        Called when the request body has been read completely. Takes care of creating the replacement body out of the
        logged parts, turning ``file`` parts into new ``data`` parts.
        """

        self._new_body = b""
        for name, part in self._parts.items():
            if "filename" in part:
                # add form fields for filename, path, size and content_type for all files contained in the request
                if "path" not in part:
                    continue

                parameters = {
                    "name": part["filename"],
                    "path": part["path"],
                    "size": str(os.stat(part["path"]).st_size),
                }
                if "content_type" in part:
                    parameters["content_type"] = part["content_type"]

                fields = {
                    self._suffixes[key]: value for (key, value) in parameters.items()
                }
                for n, p in fields.items():
                    if n is None or p is None:
                        continue
                    key = name + b"." + octoprint.util.to_bytes(n)
                    self._new_body += b"--%s\r\n" % self._multipart_boundary
                    self._new_body += (
                        b'Content-Disposition: form-data; name="%s"\r\n' % key
                    )
                    self._new_body += b"Content-Type: text/plain; charset=utf-8\r\n"
                    self._new_body += b"\r\n"
                    self._new_body += octoprint.util.to_bytes(p) + b"\r\n"
            elif "data" in part:
                self._new_body += b"--%s\r\n" % self._multipart_boundary
                value = part["data"]
                self._new_body += b'Content-Disposition: form-data; name="%s"\r\n' % name
                if "content_type" in part and part["content_type"] is not None:
                    self._new_body += b"Content-Type: %s\r\n" % part["content_type"]
                self._new_body += b"\r\n"
                self._new_body += value + b"\r\n"
        self._new_body += b"--%s--\r\n" % self._multipart_boundary

    def _handle_method(self, *args, **kwargs):
        """
        Takes care of defining the new request body if necessary and forwarding
        the current request and changed body to the ``fallback``.
        """

        # determine which body to supply
        body = b""
        if self.is_multipart():
            # make sure we really processed all data in the buffer
            while len(self._buffer):
                self._process_multipart_data(self._buffer)

            # use rewritten body
            body = self._new_body

        elif self.request.method in UploadStorageFallbackHandler.BODY_METHODS:
            # directly use data from buffer
            body = self._buffer

        # rewrite content length
        self.request.headers["Content-Length"] = len(body)

        try:
            # call the configured fallback with request and body to use
            self._fallback(self.request, body)
            self._headers_written = True
        finally:
            # make sure the temporary files are removed again
            for f in self._files:
                octoprint.util.silent_remove(f)

    # make all http methods trigger _handle_method
    get = _handle_method
    post = _handle_method
    put = _handle_method
    patch = _handle_method
    delete = _handle_method
    head = _handle_method
    options = _handle_method


def _parse_header(line, strip_quotes=True):
    parts = tornado.httputil._parseparam(";" + line)
    key = next(parts)
    pdict = {}
    for p in parts:
        i = p.find("=")
        if i >= 0:
            name = p[:i].strip().lower()
            value = p[i + 1 :].strip()
            if strip_quotes:
                value = _strip_value_quotes(value)
            pdict[name] = value
    return key, pdict


def _strip_value_quotes(value):
    if not value:
        return value

    if len(value) >= 2 and value[0] == value[-1] == '"':
        value = value[1:-1]
        value = value.replace("\\\\", "\\").replace('\\"', '"')

    return value


def _extended_header_value(value):
    if not value:
        return value

    if value.lower().startswith("iso-8859-1'") or value.lower().startswith("utf-8'"):
        # RFC 5987 section 3.2
        from urllib.parse import unquote

        encoding, _, value = value.split("'", 2)
        return unquote(value, encoding=encoding)
    else:
        # no encoding provided, strip potentially present quotes and call it a day
        return octoprint.util.to_unicode(_strip_value_quotes(value), encoding="utf-8")


class WsgiInputContainer:
    """
    A WSGI container for use with Tornado that allows supplying the request body to be used for ``wsgi.input`` in the
    generated WSGI environment upon call.

    A ``RequestHandler`` can thus provide the WSGI application with a stream for the request body, or a modified body.

    Example usage:

    .. code-block:: python

       wsgi_app = octoprint.server.util.WsgiInputContainer(octoprint_app)
       application = tornado.web.Application([
           (r".*", UploadStorageFallbackHandler, dict(fallback=wsgi_app),
       ])

    The implementation logic is basically the same as ``tornado.wsgi.WSGIContainer`` but the ``__call__`` and ``environ``
    methods have been adjusted to allow for an optionally supplied ``body`` argument which is then used for ``wsgi.input``.
    """

    def __init__(
        self, wsgi_application, headers=None, forced_headers=None, removed_headers=None
    ):
        self.wsgi_application = wsgi_application

        if headers is None:
            headers = {}
        if forced_headers is None:
            forced_headers = {}
        if removed_headers is None:
            removed_headers = []

        self.headers = headers
        self.forced_headers = forced_headers
        self.removed_headers = removed_headers

    def __call__(self, request, body=None):
        """
        Wraps the call against the WSGI app, deriving the WSGI environment from the supplied Tornado ``HTTPServerRequest``.

        :param request: the ``tornado.httpserver.HTTPServerRequest`` to derive the WSGI environment from
        :param body: an optional body  to use as ``wsgi.input`` instead of ``request.body``, can be a string or a stream
        """

        data = {}
        response = []

        def start_response(status, response_headers, exc_info=None):
            data["status"] = status
            data["headers"] = response_headers
            return response.append

        app_response = self.wsgi_application(
            WsgiInputContainer.environ(request, body), start_response
        )
        try:
            response.extend(app_response)
            body = b"".join(response)
        finally:
            if hasattr(app_response, "close"):
                app_response.close()
        if not data:
            raise Exception("WSGI app did not call start_response")

        status_code, reason = data["status"].split(" ", 1)
        status_code = int(status_code)
        headers = data["headers"]
        header_set = {k.lower() for (k, v) in headers}
        body = tornado.escape.utf8(body)
        if status_code != 304:
            if "content-length" not in header_set:
                headers.append(("Content-Length", str(len(body))))
            if "content-type" not in header_set:
                headers.append(("Content-Type", "text/html; charset=UTF-8"))

        header_set = {k.lower() for (k, v) in headers}
        for header, value in self.headers.items():
            if header.lower() not in header_set:
                headers.append((header, value))
        for header, value in self.forced_headers.items():
            headers.append((header, value))
        headers = [
            (header, value)
            for header, value in headers
            if header.lower() not in self.removed_headers
        ]

        start_line = tornado.httputil.ResponseStartLine("HTTP/1.1", status_code, reason)
        header_obj = tornado.httputil.HTTPHeaders()
        for key, value in headers:
            header_obj.add(key, value)
        request.connection.write_headers(start_line, header_obj, chunk=body)
        request.connection.finish()
        self._log(status_code, request)

    @staticmethod
    def environ(request, body=None):
        """
        Converts a ``tornado.httputil.HTTPServerRequest`` to a WSGI environment.

        An optional ``body`` to be used for populating ``wsgi.input`` can be supplied (either a string or a stream). If not
        supplied, ``request.body`` will be wrapped into a ``io.BytesIO`` stream and used instead.

        :param request: the ``tornado.httpserver.HTTPServerRequest`` to derive the WSGI environment from
        :param body: an optional body  to use as ``wsgi.input`` instead of ``request.body``, can be a string or a stream
        """
        import io

        from tornado.wsgi import to_wsgi_str

        # determine the request_body to supply as wsgi.input
        if body is not None:
            if isinstance(body, (bytes, str)):
                request_body = io.BytesIO(tornado.escape.utf8(body))
            else:
                request_body = body
        else:
            request_body = io.BytesIO(tornado.escape.utf8(request.body))

        hostport = request.host.split(":")
        if len(hostport) == 2:
            host = hostport[0]
            port = int(hostport[1])
        else:
            host = request.host
            port = 443 if request.protocol == "https" else 80
        environ = {
            "REQUEST_METHOD": request.method,
            "SCRIPT_NAME": "",
            "PATH_INFO": to_wsgi_str(
                tornado.escape.url_unescape(request.path, encoding=None, plus=False)
            ),
            "QUERY_STRING": request.query,
            "REMOTE_ADDR": request.remote_ip,
            "SERVER_NAME": host,
            "SERVER_PORT": str(port),
            "SERVER_PROTOCOL": request.version,
            "wsgi.version": (1, 0),
            "wsgi.url_scheme": request.protocol,
            "wsgi.input": request_body,
            "wsgi.errors": sys.stderr,
            "wsgi.multithread": False,
            "wsgi.multiprocess": True,
            "wsgi.run_once": False,
        }
        if "Content-Type" in request.headers:
            environ["CONTENT_TYPE"] = request.headers.pop("Content-Type")
        if "Content-Length" in request.headers:
            environ["CONTENT_LENGTH"] = request.headers.pop("Content-Length")
        for key, value in request.headers.items():
            environ["HTTP_" + key.replace("-", "_").upper()] = value
        return environ

    def _log(self, status_code, request):
        access_log = logging.getLogger("tornado.access")

        if status_code < 400:
            log_method = access_log.info
        elif status_code < 500:
            log_method = access_log.warning
        else:
            log_method = access_log.error
        request_time = 1000 * request.request_time()
        summary = request.method + " " + request.uri + " (" + request.remote_ip + ")"
        log_method("%d %s %.2fms", status_code, summary, request_time)


# ~~ customized HTTP1Connection implementation


class CustomHTTPServer(tornado.httpserver.HTTPServer):
    """
    Custom implementation of ``tornado.httpserver.HTTPServer`` that allows defining max body sizes depending on path and
    method.

    The implementation is mostly taken from ``tornado.httpserver.HTTPServer``, the only difference is the creation
    of a ``CustomHTTP1ConnectionParameters`` instance instead of ``tornado.http1connection.HTTP1ConnectionParameters``
    which is supplied with the two new constructor arguments ``max_body_sizes`` and ``max_default_body_size`` and the
    creation of a ``CustomHTTP1ServerConnection`` instead of a ``tornado.http1connection.HTTP1ServerConnection`` upon
    connection by a client.

    ``max_body_sizes`` is expected to be an iterable containing tuples of the form (method, path regex, maximum body size),
    with method and path regex having to match in order for maximum body size to take affect.

    ``default_max_body_size`` is the default maximum body size to apply if no specific one from ``max_body_sizes`` matches.
    """

    def __init__(self, *args, **kwargs):
        pass

    def initialize(self, *args, **kwargs):
        default_max_body_size = kwargs.pop("default_max_body_size", None)
        max_body_sizes = kwargs.pop("max_body_sizes", None)

        tornado.httpserver.HTTPServer.initialize(self, *args, **kwargs)

        additional = {
            "default_max_body_size": default_max_body_size,
            "max_body_sizes": max_body_sizes,
        }
        self.conn_params = CustomHTTP1ConnectionParameters.from_stock_params(
            self.conn_params, **additional
        )

    def handle_stream(self, stream, address):
        context = tornado.httpserver._HTTPRequestContext(
            stream, address, self.protocol, self.trusted_downstream
        )
        conn = CustomHTTP1ServerConnection(stream, self.conn_params, context)
        self._connections.add(conn)
        conn.start_serving(self)


class CustomHTTP1ServerConnection(tornado.http1connection.HTTP1ServerConnection):
    """
    A custom implementation of ``tornado.http1connection.HTTP1ServerConnection`` which utilizes a ``CustomHTTP1Connection``
    instead of a ``tornado.http1connection.HTTP1Connection`` in ``_server_request_loop``. The implementation logic is
    otherwise the same as ``tornado.http1connection.HTTP1ServerConnection``.
    """

    @tornado.gen.coroutine
    def _server_request_loop(self, delegate):
        try:
            while True:
                conn = CustomHTTP1Connection(
                    self.stream, False, self.params, self.context
                )
                request_delegate = delegate.start_request(self, conn)
                try:
                    ret = yield conn.read_response(request_delegate)
                except (
                    tornado.iostream.StreamClosedError,
                    tornado.iostream.UnsatisfiableReadError,
                ):
                    return
                except tornado.http1connection._QuietException:
                    # This exception was already logged.
                    conn.close()
                    return
                except Exception:
                    tornado.http1connection.gen_log.error(
                        "Uncaught exception", exc_info=True
                    )
                    conn.close()
                    return
                if not ret:
                    return
                yield tornado.gen.moment
        finally:
            delegate.on_close(self)


class CustomHTTP1Connection(tornado.http1connection.HTTP1Connection):
    """
    A custom implementation of ``tornado.http1connection.HTTP1Connection`` which upon checking the ``Content-Length`` of
    the request against the configured maximum utilizes ``max_body_sizes`` and ``default_max_body_size`` as a fallback.
    """

    def __init__(self, stream, is_client, params=None, context=None):
        if params is None:
            params = CustomHTTP1ConnectionParameters()

        tornado.http1connection.HTTP1Connection.__init__(
            self, stream, is_client, params=params, context=context
        )

        import re

        self._max_body_sizes = list(
            map(
                lambda x: (x[0], re.compile(x[1]), x[2]),
                self.params.max_body_sizes or list(),
            )
        )
        self._default_max_body_size = (
            self.params.default_max_body_size or self.stream.max_buffer_size
        )

    def _read_body(self, code, headers, delegate):
        """
        Basically the same as ``tornado.http1connection.HTTP1Connection._read_body``, but determines the maximum
        content length individually for the request (utilizing ``._get_max_content_length``).

        If the individual max content length is 0 or smaller no content length is checked. If the content length of the
        current request exceeds the individual max content length, the request processing is aborted and an
        ``HTTPInputError`` is raised.
        """
        if "Content-Length" in headers:
            if "Transfer-Encoding" in headers:
                # Response cannot contain both Content-Length and
                # Transfer-Encoding headers.
                # http://tools.ietf.org/html/rfc7230#section-3.3.3
                raise tornado.httputil.HTTPInputError(
                    "Response with both Transfer-Encoding and Content-Length"
                )
            if "," in headers["Content-Length"]:
                # Proxies sometimes cause Content-Length headers to get
                # duplicated.  If all the values are identical then we can
                # use them but if they differ it's an error.
                pieces = re.split(r",\s*", headers["Content-Length"])
                if any(i != pieces[0] for i in pieces):
                    raise tornado.httputil.HTTPInputError(
                        "Multiple unequal Content-Lengths: %r" % headers["Content-Length"]
                    )
                headers["Content-Length"] = pieces[0]

            try:
                content_length = int(headers["Content-Length"])
            except ValueError:
                # Handles non-integer Content-Length value.
                raise tornado.httputil.HTTPInputError(
                    "Only integer Content-Length is allowed: %s"
                    % headers["Content-Length"]
                )

            max_content_length = self._get_max_content_length(
                self._request_start_line.method, self._request_start_line.path
            )
            if (
                max_content_length is not None
                and 0 <= max_content_length < content_length
            ):
                raise tornado.httputil.HTTPInputError("Content-Length too long")
        else:
            content_length = None

        if code == 204:
            # This response code is not allowed to have a non-empty body,
            # and has an implicit length of zero instead of read-until-close.
            # http://www.w3.org/Protocols/rfc2616/rfc2616-sec4.html#sec4.3
            if "Transfer-Encoding" in headers or content_length not in (None, 0):
                raise tornado.httputil.HTTPInputError(
                    "Response with code %d should not have body" % code
                )
            content_length = 0

        if content_length is not None:
            return self._read_fixed_body(content_length, delegate)
        if headers.get("Transfer-Encoding") == "chunked":
            return self._read_chunked_body(delegate)
        if self.is_client:
            return self._read_body_until_close(delegate)
        return None

    def _get_max_content_length(self, method, path):
        """
        Gets the max content length for the given method and path. Checks whether method and path match against any
        of the specific maximum content lengths supplied in ``max_body_sizes`` and returns that as the maximum content
        length if available, otherwise returns ``default_max_body_size``.

        :param method: method of the request to match against
        :param path: path of the request to match against
        :return: determine maximum content length to apply to this request, max return 0 for unlimited allowed content
                 length
        """

        for m, p, s in self._max_body_sizes:
            if method == m and p.match(path):
                return s
        return self._default_max_body_size


class CustomHTTP1ConnectionParameters(tornado.http1connection.HTTP1ConnectionParameters):
    """
    An implementation of ``tornado.http1connection.HTTP1ConnectionParameters`` that adds two new parameters
    ``max_body_sizes`` and ``default_max_body_size``.

    For a description of these please see the documentation of ``CustomHTTPServer`` above.
    """

    def __init__(self, *args, **kwargs):
        max_body_sizes = kwargs.pop("max_body_sizes", list())
        default_max_body_size = kwargs.pop("default_max_body_size", None)

        tornado.http1connection.HTTP1ConnectionParameters.__init__(self, *args, **kwargs)

        self.max_body_sizes = max_body_sizes
        self.default_max_body_size = default_max_body_size

    @classmethod
    def from_stock_params(cls, other, **additional):
        kwargs = dict(other.__dict__)
        for key, value in additional.items():
            kwargs[key] = value
        return cls(**kwargs)


# ~~ customized large response handler


class LargeResponseHandler(
    RequestlessExceptionLoggingMixin, CorsSupportMixin, tornado.web.StaticFileHandler
):
    """
    Customized `tornado.web.StaticFileHandler <http://tornado.readthedocs.org/en/branch4.0/web.html#tornado.web.StaticFileHandler>`_
    that allows delivery of the requested resource as attachment and access and request path validation through
    optional callbacks. Note that access validation takes place before path validation.

    Arguments:
       path (str): The system path from which to serve files (this will be forwarded to the ``initialize`` method of
           :class:``~tornado.web.StaticFileHandler``)
       default_filename (str): The default filename to serve if none is explicitly specified and the request references
           a subdirectory of the served path (this will be forwarded to the ``initialize`` method of
           :class:``~tornado.web.StaticFileHandler`` as the ``default_filename`` keyword parameter). Defaults to ``None``.
       as_attachment (bool): Whether to serve requested files with ``Content-Disposition: attachment`` header (``True``)
           or not. Defaults to ``False``.
       allow_client_caching (bool): Whether to allow the client to cache (by not setting any ``Cache-Control`` or
           ``Expires`` headers on the response) or not.
       access_validation (function): Callback to call in the ``get`` method to validate access to the resource. Will
           be called with ``self.request`` as parameter which contains the full tornado request object. Should raise
           a ``tornado.web.HTTPError`` if access is not allowed in which case the request will not be further processed.
           Defaults to ``None`` and hence no access validation being performed.
       path_validation (function): Callback to call in the ``get`` method to validate the requested path. Will be called
           with the requested path as parameter. Should raise a ``tornado.web.HTTPError`` (e.g. an 404) if the requested
           path does not pass validation in which case the request will not be further processed.
           Defaults to ``None`` and hence no path validation being performed.
       etag_generator (function): Callback to call for generating the value of the ETag response header. Will be
           called with the response handler as parameter. May return ``None`` to prevent the ETag response header
           from being set. If not provided the last modified time of the file in question will be used as returned
           by ``get_content_version``.
       name_generator (function): Callback to call for generating the value of the attachment file name header. Will be
           called with the requested path as parameter.
       mime_type_guesser (function): Callback to guess the mime type to use for the content type encoding of the
           response. Will be called with the requested path on disk as parameter.
       is_pre_compressed (bool): if the file is expected to be pre-compressed, i.e, if there is a file in the same
           directory with the same name, but with '.gz' appended and gzip-encoded
    """

    def initialize(
        self,
        path,
        default_filename=None,
        as_attachment=False,
        allow_client_caching=True,
        access_validation=None,
        path_validation=None,
        etag_generator=None,
        name_generator=None,
        mime_type_guesser=None,
        is_pre_compressed=False,
        stream_body=False,
    ):
        tornado.web.StaticFileHandler.initialize(
            self, os.path.abspath(path), default_filename
        )
        self._as_attachment = as_attachment
        self._allow_client_caching = allow_client_caching
        self._access_validation = access_validation
        self._path_validation = path_validation
        self._etag_generator = etag_generator
        self._name_generator = name_generator
        self._mime_type_guesser = mime_type_guesser
        self._is_pre_compressed = is_pre_compressed
        self._stream_body = stream_body

    def should_use_precompressed(self):
        return self._is_pre_compressed and "gzip" in self.request.headers.get(
            "Accept-Encoding", ""
        )

    def get(self, path, include_body=True):
        if self._access_validation is not None:
            self._access_validation(self.request)
        if self._path_validation is not None:
            self._path_validation(path)

        if "cookie" in self.request.arguments:
            self.set_cookie(self.request.arguments["cookie"][0], "true", path="/")

        if self.should_use_precompressed():
            if os.path.exists(os.path.join(self.root, path + ".gz")):
                self.set_header("Content-Encoding", "gzip")
                path = path + ".gz"
            else:
                logging.getLogger(__name__).warning(
                    "Precompressed assets expected but {}.gz does not exist "
                    "in {}, using plain file instead.".format(path, self.root)
                )

        if self._stream_body:
            return self.streamed_get(path, include_body=include_body)
        else:
            return tornado.web.StaticFileHandler.get(
                self, path, include_body=include_body
            )

    @tornado.gen.coroutine
    def streamed_get(self, path, include_body=True):
        """
        Version of StaticFileHandler.get that doesn't support ranges or ETag but streams the content. Helpful for files
        that might still change while being transmitted (e.g. log files)
        """

        # Set up our path instance variables.
        self.path = self.parse_url_path(path)
        del path  # make sure we don't refer to path instead of self.path again
        absolute_path = self.get_absolute_path(self.root, self.path)
        self.absolute_path = self.validate_absolute_path(self.root, absolute_path)
        if self.absolute_path is None:
            return

        content_type = self.get_content_type()
        if content_type:
            self.set_header("Content-Type", content_type)
        self.set_extra_headers(self.path)

        if include_body:
            content = self.get_content(self.absolute_path)
            if isinstance(content, bytes):
                content = [content]
            for chunk in content:
                try:
                    self.write(chunk)
                    yield self.flush()
                except tornado.iostream.StreamClosedError:
                    return
        else:
            assert self.request.method == "HEAD"

    def set_extra_headers(self, path):
        if self._as_attachment:
            filename = None
            if callable(self._name_generator):
                filename = self._name_generator(path)
            if filename is None:
                filename = os.path.basename(path)

            filename = tornado.escape.url_escape(filename, plus=False)
            self.set_header(
                "Content-Disposition",
                "attachment; filename=\"{}\"; filename*=UTF-8''{}".format(
                    filename, filename
                ),
            )

        if not self._allow_client_caching:
            self.set_header("Cache-Control", "max-age=0, must-revalidate, private")
            self.set_header("Expires", "-1")

    @property
    def original_absolute_path(self):
        """The path of the uncompressed file corresponding to the compressed file"""
        if self._is_pre_compressed:
            return self.absolute_path.rstrip(".gz")
        return self.absolute_path

    def compute_etag(self):
        if self._etag_generator is not None:
            etag = self._etag_generator(self)
        else:
            etag = str(self.get_content_version(self.absolute_path))

        if not etag.endswith('"'):
            etag = f'"{etag}"'
        return etag

    # noinspection PyAttributeOutsideInit
    def get_content_type(self):
        if self._mime_type_guesser is not None:
            type = self._mime_type_guesser(self.original_absolute_path)
            if type is not None:
                return type

        correct_absolute_path = None
        try:
            # reset self.absolute_path temporarily
            if self.should_use_precompressed():
                correct_absolute_path = self.absolute_path
                self.absolute_path = self.original_absolute_path
            return tornado.web.StaticFileHandler.get_content_type(self)
        finally:
            # restore self.absolute_path
            if self.should_use_precompressed() and correct_absolute_path is not None:
                self.absolute_path = correct_absolute_path

    @classmethod
    def get_content_version(cls, abspath):
        import os
        import stat

        return os.stat(abspath)[stat.ST_MTIME]


##~~ URL Forward Handler for forwarding requests to a preconfigured static URL


class UrlProxyHandler(
    RequestlessExceptionLoggingMixin, CorsSupportMixin, tornado.web.RequestHandler
):
    """
    `tornado.web.RequestHandler <http://tornado.readthedocs.org/en/branch4.0/web.html#request-handlers>`_ that proxies
    requests to a preconfigured url and returns the response. Allows delivery of the requested content as attachment
    and access validation through an optional callback.

    This will use `tornado.httpclient.AsyncHTTPClient <http://tornado.readthedocs.org/en/branch4.0/httpclient.html#tornado.httpclient.AsyncHTTPClient>`_
    for making the request to the configured endpoint and return the body of the client response with the status code
    from the client response and the following headers:

      * ``Date``, ``Cache-Control``, ``Expires``, ``ETag``, ``Server``, ``Content-Type`` and ``Location`` will be copied over.
      * If ``as_attachment`` is set to True, ``Content-Disposition`` will be set to ``attachment``. If ``basename`` is
        set including the attachment's ``filename`` attribute will be set to the base name followed by the extension
        guessed based on the MIME type from the ``Content-Type`` header of the response. If no extension can be guessed
        no ``filename`` attribute will be set.

    Arguments:
       url (str): URL to forward any requests to. A 404 response will be returned if this is not set. Defaults to ``None``.
       as_attachment (bool): Whether to serve files with ``Content-Disposition: attachment`` header (``True``)
           or not. Defaults to ``False``.
       basename (str): base name of file names to return as part of the attachment header, see above. Defaults to ``None``.
       access_validation (function): Callback to call in the ``get`` method to validate access to the resource. Will
           be called with ``self.request`` as parameter which contains the full tornado request object. Should raise
           a ``tornado.web.HTTPError`` if access is not allowed in which case the request will not be further processed.
           Defaults to ``None`` and hence no access validation being performed.
    """

    def initialize(
        self, url=None, as_attachment=False, basename=None, access_validation=None
    ):
        tornado.web.RequestHandler.initialize(self)
        self._url = url
        self._as_attachment = as_attachment
        self._basename = basename
        self._access_validation = access_validation

    @tornado.gen.coroutine
    def get(self, *args, **kwargs):
        if self._access_validation is not None:
            self._access_validation(self.request)

        if self._url is None:
            raise tornado.web.HTTPError(404)

        client = tornado.httpclient.AsyncHTTPClient()
        r = tornado.httpclient.HTTPRequest(
            url=self._url,
            method=self.request.method,
            body=self.request.body,
            headers=self.request.headers,
            follow_redirects=False,
            allow_nonstandard_methods=True,
        )

        try:
            return client.fetch(r, self.handle_response)
        except tornado.web.HTTPError as e:
            if hasattr(e, "response") and e.response:
                self.handle_response(e.response)
            else:
                raise tornado.web.HTTPError(500)

    def handle_response(self, response):
        if response.error and not isinstance(response.error, tornado.web.HTTPError):
            raise tornado.web.HTTPError(500)

        filename = None

        self.set_status(response.code)
        for name in (
            "Date",
            "Cache-Control",
            "Server",
            "Content-Type",
            "Location",
            "Expires",
            "ETag",
        ):
            value = response.headers.get(name)
            if value:
                self.set_header(name, value)

                if name == "Content-Type":
                    filename = self.get_filename(value)

        if self._as_attachment:
            if filename is not None:
                self.set_header(
                    "Content-Disposition", "attachment; filename=%s" % filename
                )
            else:
                self.set_header("Content-Disposition", "attachment")

        if response.body:
            self.write(response.body)
        self.finish()

    def get_filename(self, content_type):
        if not self._basename:
            return None

        typeValue = list(x.strip() for x in content_type.split(";"))
        if len(typeValue) == 0:
            return None

        extension = mimetypes.guess_extension(typeValue[0])
        if not extension:
            return None

        return f"{self._basename}{extension}"


class StaticDataHandler(
    RequestlessExceptionLoggingMixin, CorsSupportMixin, tornado.web.RequestHandler
):
    """
    `tornado.web.RequestHandler <http://tornado.readthedocs.org/en/branch4.0/web.html#request-handlers>`_ that returns
    static ``data`` of a configured ``content_type``.

    Arguments:
       data (str): The data with which to respond
       content_type (str): The content type with which to respond. Defaults to ``text/plain``
    """

    def initialize(self, data="", content_type="text/plain"):
        self.data = data
        self.content_type = content_type

    def get(self, *args, **kwargs):
        self.set_status(200)
        self.set_header("Content-Type", self.content_type)
        self.write(self.data)
        self.flush()
        self.finish()


class DeprecatedEndpointHandler(CorsSupportMixin, tornado.web.RequestHandler):
    """
    `tornado.web.RequestHandler <http://tornado.readthedocs.org/en/branch4.0/web.html#request-handlers>`_ that redirects
    to another ``url`` and logs a deprecation warning.

    Arguments:
       url (str): URL to which to redirect
    """

    def initialize(self, url):
        self._url = url
        self._logger = logging.getLogger(__name__)

    def _handle_method(self, *args, **kwargs):
        to_url = self._url.format(*args)
        self._logger.info(
            f"Redirecting deprecated endpoint {self.request.path} to {to_url}"
        )
        self.redirect(to_url, permanent=True)

    # make all http methods trigger _handle_method
    get = _handle_method
    post = _handle_method
    put = _handle_method
    patch = _handle_method
    delete = _handle_method
    head = _handle_method
    options = _handle_method


class StaticZipBundleHandler(CorsSupportMixin, tornado.web.RequestHandler):
    def initialize(
        self,
        files=None,
        as_attachment=True,
        attachment_name=None,
        access_validation=None,
        compress=False,
    ):
        if files is None:
            files = []
        if as_attachment and not attachment_name:
            raise ValueError("attachment name must be set if as_attachment is True")

        self._files = files
        self._as_attachment = as_attachment
        self._attachment_name = attachment_name
        self._access_validator = access_validation
        self._compress = compress

    def get(self, *args, **kwargs):
        if self._access_validator is not None:
            self._access_validator(self.request)
        return self.stream_zip(self._files)

    def get_attachment_name(self):
        return self._attachment_name

    def normalize_files(self, files):
        result = []
        for f in files:
            if isinstance(f, str):
                result.append({"path": f})
            elif isinstance(f, dict) and ("path" in f or "iter" in f or "content" in f):
                result.append(f)
        return result

    @tornado.gen.coroutine
    def stream_zip(self, files):
        self.set_header("Content-Type", "application/zip")
        if self._as_attachment:
            self.set_header(
                "Content-Disposition",
                f'attachment; filename="{self.get_attachment_name()}"',
            )

        z = ZipStream(sized=True)
        if self._compress:
            try:
                z = ZipStream(compress_type=ZIP_DEFLATED)
            except RuntimeError:
                # no zlib support
                pass

        for f in self.normalize_files(files):
            name = f.get("name")
            path = f.get("path")
            data = f.get("iter") or f.get("content")

            if path:
                z.add_path(path, arcname=name)
            elif data and name:
                z.add(data, arcname=name)

        if z.sized:
            self.set_header("Content-Length", len(z))
        self.set_header("Last-Modified", z.last_modified)

        for chunk in z:
            try:
                self.write(chunk)
                yield self.flush()
            except tornado.iostream.StreamClosedError:
                return


class DynamicZipBundleHandler(StaticZipBundleHandler):
    # noinspection PyMethodOverriding
    def initialize(
        self,
        path_validation=None,
        path_processor=None,
        as_attachment=True,
        attachment_name=None,
        access_validation=None,
        compress=False,
    ):
        if as_attachment and not attachment_name:
            raise ValueError("attachment name must be set if as_attachment is True")

        self._path_validator = path_validation
        self._path_processor = path_processor
        self._as_attachment = as_attachment
        self._attachment_name = attachment_name
        self._access_validator = access_validation
        self._compress = compress

    def get(self, *args, **kwargs):
        if self._access_validator is not None:
            self._access_validator(self.request)

        files = list(
            map(octoprint.util.to_unicode, self.request.query_arguments.get("files", []))
        )
        return self._get_files_zip(files)

    def post(self, *args, **kwargs):
        if self._access_validator is not None:
            self._access_validator(self.request)

        import json

        content_type = self.request.headers.get("Content-Type", "")
        try:
            if "application/json" in content_type:
                data = json.loads(self.request.body)
            else:
                data = self.request.body_arguments
        except Exception:
            raise tornado.web.HTTPError(400)

        return self._get_files_zip(
            list(map(octoprint.util.to_unicode, data.get("files", [])))
        )

    def _get_files_zip(self, files):
        files = self.normalize_files(files)
        if not files:
            raise tornado.web.HTTPError(400)

        for f in files:
            if "path" in f:
                if callable(self._path_processor):
                    path = self._path_processor(f["path"])
                    if isinstance(path, tuple):
                        f["name"], f["path"] = path
                    else:
                        f["path"] = path
                self._path_validator(f["path"])

        return self.stream_zip(files)


class SystemInfoBundleHandler(CorsSupportMixin, tornado.web.RequestHandler):
    # noinspection PyMethodOverriding
    def initialize(self, access_validation=None):
        self._access_validator = access_validation

    @tornado.gen.coroutine
    def get(self, *args, **kwargs):
        if self._access_validator is not None:
            self._access_validator(self.request)

        from octoprint.cli.systeminfo import (
            get_systeminfo,
            get_systeminfo_bundle,
            get_systeminfo_bundle_name,
        )
        from octoprint.server import (
            connectivityChecker,
            environmentDetector,
            pluginManager,
            printer,
            safe_mode,
        )
        from octoprint.settings import settings

        systeminfo = get_systeminfo(
            environmentDetector,
            connectivityChecker,
            settings(),
            {
                "browser.user_agent": self.request.headers.get("User-Agent"),
                "octoprint.safe_mode": safe_mode is not None,
                "systeminfo.generator": "zipapi",
            },
        )

        z = get_systeminfo_bundle(
            systeminfo,
            settings().getBaseFolder("logs"),
            printer=printer,
            plugin_manager=pluginManager,
        )

        self.set_header("Content-Type", "application/zip")
        self.set_header(
            "Content-Disposition",
            f'attachment; filename="{get_systeminfo_bundle_name()}"',
        )
        if z.sized:
            self.set_header("Content-Length", len(z))
        self.set_header("Last-Modified", z.last_modified)

        for chunk in z:
            try:
                self.write(chunk)
                yield self.flush()
            except tornado.iostream.StreamClosedError:
                return

    def get_attachment_name(self):
        import time

        return "octoprint-systeminfo-{}.zip".format(time.strftime("%Y%m%d%H%M%S"))


class GlobalHeaderTransform(tornado.web.OutputTransform):

    HEADERS = {}
    FORCED_HEADERS = {}
    REMOVED_HEADERS = []

    @classmethod
    def for_headers(cls, name, headers=None, forced_headers=None, removed_headers=None):
        if headers is None:
            headers = {}
        if forced_headers is None:
            forced_headers = {}
        if removed_headers is None:
            removed_headers = []

        return type(
            name,
            (GlobalHeaderTransform,),
            {
                "HEADERS": headers,
                "FORCED_HEADERS": forced_headers,
                "REMOVED_HEADERS": removed_headers,
            },
        )

    def __init__(self, request):
        tornado.web.OutputTransform.__init__(self, request)

    def transform_first_chunk(self, status_code, headers, chunk, finishing):
        for header, value in self.HEADERS.items():
            if header not in headers:
                headers[header] = value
        for header, value in self.FORCED_HEADERS.items():
            headers[header] = value
        for header in self.REMOVED_HEADERS:
            del headers[header]
        return status_code, headers, chunk


# ~~ Factory method for creating Flask access validation wrappers from the Tornado request context


def access_validation_factory(app, validator, *args):
    """
    Creates an access validation wrapper using the supplied validator.

    :param validator: the access validator to use inside the validation wrapper
    :return: an access validator taking a request as parameter and performing the request validation
    """

    # noinspection PyProtectedMember
    def f(request):
        """
        Creates a custom wsgi and Flask request context in order to be able to process user information
        stored in the current session.

        :param request: The Tornado request for which to create the environment and context
        """
        import flask

        wsgi_environ = WsgiInputContainer.environ(request)
        with app.request_context(wsgi_environ):
            session = app.session_interface.open_session(app, flask.request)
            user_id = session.get("_user_id")
            user = None

            # Yes, using protected methods is ugly. But these used to be publicly available in former versions
            # of flask-login, there are no replacements, and seeing them renamed & hidden in a minor version release
            # without any mention in the changelog means the public API ain't strictly stable either, so we might
            # as well make our life easier here and just use them...
            if user_id is not None and app.login_manager._user_callback is not None:
                user = app.login_manager._user_callback(user_id)
            app.login_manager._update_request_context_with_user(user)

            validator(flask.request, *args)

    return f


def path_validation_factory(path_filter, status_code=404):
    """
    Creates a request path validation wrapper returning the defined status code if the supplied path_filter returns False.

    :param path_filter: the path filter to use on the requested path, should return False for requests that should
       be responded with the provided error code.
    :return: a request path validator taking a request path as parameter and performing the request validation
    """

    def f(path):
        if not path_filter(path):
            raise tornado.web.HTTPError(status_code)

    return f


def validation_chain(*validators):
    def f(request):
        for validator in validators:
            validator(request)

    return f



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2020 The OctoPrint Project - Released under terms of the AGPLv3 License"

import time

import flask
from flask_babel import gettext

import octoprint.plugin
from octoprint.access import USER_GROUP
from octoprint.access.permissions import Permissions
from octoprint.events import Events


class ActionCommandNotificationPlugin(
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.SettingsPlugin,
    octoprint.plugin.SimpleApiPlugin,
    octoprint.plugin.TemplatePlugin,
    octoprint.plugin.EventHandlerPlugin,
):
    def __init__(self):
        self._notifications = []

    # Additional permissions hook

    def get_additional_permissions(self):
        return [
            {
                "key": "SHOW",
                "name": "Show printer notifications",
                "description": gettext("Allows to see printer notifications"),
                "default_groups": [USER_GROUP],
                "roles": ["show"],
            },
            {
                "key": "CLEAR",
                "name": "Clear printer notifications",
                "description": gettext("Allows to clear printer notifications"),
                "default_groups": [USER_GROUP],
                "roles": ["clear"],
            },
        ]

    # ~ AssetPlugin

    def get_assets(self):
        return {
            "js": ["js/action_command_notification.js"],
            "clientjs": ["clientjs/action_command_notification.js"],
            "css": ["css/action_command_notification.css"],
        }

    # ~ EventHandlerPlugin

    def on_event(self, event, payload):
        if event == Events.DISCONNECTED:
            self._clear_notifications()

    # ~ SettingsPlugin

    def get_settings_defaults(self):
        return {"enable": True, "enable_popups": False}

    # ~ SimpleApiPlugin

    def on_api_get(self, request):
        if not Permissions.PLUGIN_ACTION_COMMAND_NOTIFICATION_SHOW.can():
            return flask.abort(403)
        return flask.jsonify(
            notifications=[
                {"timestamp": notification[0], "message": notification[1]}
                for notification in self._notifications
            ]
        )

    def get_api_commands(self):
        return {"clear": []}

    def on_api_command(self, command, data):
        if command == "clear":
            if not Permissions.PLUGIN_ACTION_COMMAND_NOTIFICATION_CLEAR.can():
                return flask.abort(403, "Insufficient permissions")
            self._clear_notifications()

    # ~ TemplatePlugin

    def get_template_configs(self):
        return [
            {
                "type": "settings",
                "name": gettext("Printer Notifications"),
                "custom_bindings": False,
            },
            {
                "type": "sidebar",
                "name": gettext("Printer Notifications"),
                "icon": "bell-o",
                "styles_wrapper": ["display: none"],
                "template_header": "action_command_notification_sidebar_header.jinja2",
                "data_bind": "visible: loginState.hasPermissionKo(access.permissions.PLUGIN_ACTION_COMMAND_NOTIFICATION_SHOW)"
                " && settings.settings.plugins.action_command_notification.enable()",
            },
        ]

    # ~ action command handler

    def action_command_handler(self, comm, line, action, *args, **kwargs):
        if not self._settings.get_boolean(["enable"]):
            return

        parts = action.split(None, 1)
        if len(parts) == 1:
            action = parts[0]
            parameter = ""
        else:
            action, parameter = parts

        if action != "notification":
            return

        message = parameter.strip()
        self._notifications.append((time.time(), message))
        self._plugin_manager.send_plugin_message(self._identifier, {"message": message})

        self._logger.info(f"Got a notification: {message}")

    def _clear_notifications(self):
        self._notifications = []
        self._plugin_manager.send_plugin_message(self._identifier, {})
        self._logger.info("Notifications cleared")


__plugin_name__ = "Action Command Notification Support"
__plugin_description__ = (
    "Allows your printer to trigger notifications via action commands on the connection"
)
__plugin_author__ = "Gina Häußge"
__plugin_disabling_discouraged__ = gettext(
    "Without this plugin your printer will no longer be able to trigger"
    " notifications in OctoPrint"
)
__plugin_license__ = "AGPLv3"
__plugin_pythoncompat__ = ">=3.7,<4"
__plugin_implementation__ = ActionCommandNotificationPlugin()
__plugin_hooks__ = {
    "octoprint.comm.protocol.action": __plugin_implementation__.action_command_handler,
    "octoprint.access.permissions": __plugin_implementation__.get_additional_permissions,
}



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"


from importlib import import_module

import click
from werkzeug.utils import cached_property


class LazyGroup(click.Group):
    """
    A click Group that imports the actual implementation only when
    needed.  This allows for more resilient CLIs where the top-level
    command does not fail when a subcommand is broken enough to fail
    at import time.
    """

    # shared by https://github.com/indico/indico in pallets/click#945

    def __init__(self, import_name, **kwargs):
        self._import_name = import_name
        click.Group.__init__(self, **kwargs)

    @cached_property
    def _impl(self):
        module, name = self._import_name.split(":", 1)
        return getattr(import_module(module), name)

    def get_command(self, ctx, cmd_name):
        return self._impl.get_command(ctx, cmd_name)

    def list_commands(self, ctx):
        return self._impl.list_commands(ctx)

    def invoke(self, ctx):
        return self._impl.invoke(ctx)

    def get_usage(self, ctx):
        return self._impl.get_usage(ctx)

    def get_params(self, ctx):
        return self._impl.get_params(ctx)



import datetime

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import copy
import hashlib
import logging
import logging.handlers
import os
import threading
import time
from concurrent import futures

import flask
import requests
from flask_babel import gettext

import octoprint.plugin
import octoprint.settings
from octoprint.access import ADMIN_GROUP, USER_GROUP
from octoprint.access.permissions import Permissions
from octoprint.server import BRANCH, NO_CONTENT, REVISION, VERSION
from octoprint.server.util.flask import (
    check_etag,
    no_firstrun_access,
    with_revalidation_checking,
)
from octoprint.util import RepeatedTimer, dict_merge, get_formatted_size, to_unicode, yaml
from octoprint.util.commandline import CommandlineError
from octoprint.util.pip import create_pip_caller
from octoprint.util.version import (
    get_comparable_version,
    get_python_version_string,
    is_python_compatible,
    is_released_octoprint_version,
    is_stable,
)

from . import cli, exceptions, updaters, util, version_checks

# OctoPi 0.16+
MINIMUM_PYTHON = "2.7.13"
MINIMUM_SETUPTOOLS = "40.7.1"
MINIMUM_PIP = "19.0.1"

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"


##~~ Plugin


class SoftwareUpdatePlugin(
    octoprint.plugin.BlueprintPlugin,
    octoprint.plugin.SettingsPlugin,
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.TemplatePlugin,
    octoprint.plugin.StartupPlugin,
    octoprint.plugin.WizardPlugin,
    octoprint.plugin.EventHandlerPlugin,
):
    COMMIT_TRACKING_TYPES = ("github_commit", "bitbucket_commit")
    CURRENT_TRACKING_TYPES = COMMIT_TRACKING_TYPES + ("etag", "lastmodified", "jsondata")
    RELEASE_TRACKING_TYPES = ("github_release",)

    OCTOPRINT_RESTART_TYPES = ("pip", "single_file_plugin")

    VALID_RESTART_TYPES = ("octoprint", "environment")

    DATA_FORMAT_VERSION = "v4"

    CHECK_OVERLAY_KEY = "softwareupdate_check_overlay"

    # noinspection PyMissingConstructor
    def __init__(self):
        self._update_in_progress = False
        self._configured_checks_mutex = threading.Lock()
        self._configured_checks = None
        self._refresh_configured_checks = False

        self._get_versions_mutex = threading.RLock()
        self._get_versions_data = None
        self._get_versions_data_ready = threading.Event()

        self._version_cache = {}
        self._version_cache_ttl = 0
        self._version_cache_path = None
        self._version_cache_dirty = False
        self._version_cache_timestamp = None

        self._overlay_cache = {}
        self._overlay_cache_ttl = 0
        self._overlay_cache_timestamp = None

        self._update_log = []
        self._update_log_path = None
        self._update_log_dirty = False
        self._update_log_mutex = threading.RLock()
        self._queued_updates = {"targets": [], "force": True}
        self._queued_updates_abort_timer = None
        self._print_cancelled = False

        self._environment_supported = True
        self._environment_versions = {}
        self._environment_ready = threading.Event()

        self._storage_sufficient = True
        self._storage_info = {}

        self._console_logger = None

        self._get_throttled = lambda: False

    def initialize(self):
        self._console_logger = logging.getLogger(
            "octoprint.plugins.softwareupdate.console"
        )

        self._version_cache_ttl = self._settings.get_int(["cache_ttl"]) * 60
        self._version_cache_path = os.path.join(
            self.get_plugin_data_folder(), "versioncache.yaml"
        )
        self._load_version_cache()

        self._update_log_path = os.path.join(
            self.get_plugin_data_folder(), "updatelog.yaml"
        )
        self._load_update_log()

        self._overlay_cache_ttl = self._settings.get_int(["check_overlay_ttl"]) * 60

        def refresh_checks(name, plugin):
            self._refresh_configured_checks = True
            self._send_client_message("update_versions")

        self._plugin_lifecycle_manager.add_callback("enabled", refresh_checks)
        self._plugin_lifecycle_manager.add_callback("disabled", refresh_checks)

    # Additional permissions hook

    def get_additional_permissions(self):
        return [
            {
                "key": "CHECK",
                "name": "Check",
                "description": gettext("Allows to check for software updates"),
                "roles": ["check"],
                "default_groups": [USER_GROUP],
            },
            {
                "key": "UPDATE",
                "name": "Update",
                "description": gettext("Allows to perform software updates"),
                "default_groups": [ADMIN_GROUP],
                "roles": ["update"],
                "dangerous": True,
            },
            {
                "key": "CONFIGURE",
                "name": "Configure",
                "description": gettext("Allows to configure software update"),
                "default_groups": [ADMIN_GROUP],
                "roles": ["configure"],
                "dangerous": True,
            },
        ]

    # Additional bundle contents

    def get_additional_bundle_files(self, *args, **kwargs):
        console_log = self._settings.get_plugin_logfile_path(postfix="console")

        def updatelog():
            if not self._update_log:
                self._load_update_log()

            result = []
            for entry in self._update_log:
                line = "{datetime} - {display_name} - {current_version} -> {target_version} - {success} - {text}".format(
                    datetime=entry["datetime"].replace("T", " "),
                    display_name=entry["display_name"],
                    current_version=entry["current_version"],
                    target_version=entry["target_version"],
                    success="SUCCESS" if entry["success"] else "FAILED",
                    text=entry["text"]
                    + (
                        " (Release notes: {})".format(entry["release_notes"])
                        if entry["release_notes"]
                        else ""
                    ),
                )
                result.append(line)
            return "\n".join(result)

        return {
            os.path.basename(console_log): console_log,
            "plugin_softwareupdate_update.log": updatelog,
        }

    def on_startup(self, host, port):
        console_logging_handler = logging.handlers.RotatingFileHandler(
            self._settings.get_plugin_logfile_path(postfix="console"),
            maxBytes=2 * 1024 * 1024,
            encoding="utf-8",
        )
        console_logging_handler.setFormatter(logging.Formatter("%(asctime)s %(message)s"))
        console_logging_handler.setLevel(logging.DEBUG)

        self._console_logger.addHandler(console_logging_handler)
        self._console_logger.setLevel(logging.DEBUG)
        self._console_logger.propagate = False

        helpers = self._plugin_manager.get_helpers("pi_support", "get_throttled")
        if helpers and "get_throttled" in helpers:
            self._get_throttled = helpers["get_throttled"]
            if self._settings.get_boolean(["ignore_throttled"]):
                self._logger.warning(
                    "!!! THROTTLE STATE IGNORED !!! You have configured the Software Update plugin to ignore an active throttle state of the underlying system. You might run into stability issues or outright corrupt your install. Consider fixing the throttling issue instead of suppressing it."
                )

    def on_after_startup(self):
        self._check_environment()
        self._check_storage()
        self.get_current_versions()

    def _get_configured_checks(self):
        with self._configured_checks_mutex:
            if (
                self._refresh_configured_checks
                or self._configured_checks is None
                or self._check_overlays_stale
            ):

                self._refresh_configured_checks = False

                overlays = self._get_check_overlays()
                self._configured_checks = self._settings.get(["checks"], merged=True)

                update_check_hooks = self._plugin_manager.get_hooks(
                    "octoprint.plugin.softwareupdate.check_config"
                )
                check_providers = {}

                effective_configs = {}

                for name, hook in update_check_hooks.items():
                    try:
                        hook_checks = hook()
                    except Exception:
                        self._logger.exception(
                            f"Error while retrieving update information "
                            f"from plugin {name}",
                            extra={"plugin": name},
                        )
                    else:
                        for key, default_config in hook_checks.items():
                            if key in effective_configs or key == "octoprint":
                                if key == name:
                                    self._logger.warning(
                                        "Software update hook {} provides check for itself but that was already registered by {} - overwriting that third party registration now!".format(
                                            name, check_providers.get(key, "unknown hook")
                                        )
                                    )
                                else:
                                    self._logger.warning(
                                        "Software update hook {} tried to overwrite config for check {} but that was already configured elsewhere".format(
                                            name, key
                                        )
                                    )
                                    continue

                            check_providers[key] = name

                            yaml_config = {}
                            effective_config = copy.deepcopy(default_config)

                            if key in overlays:
                                effective_config = dict_merge(
                                    default_config, overlays[key]
                                )

                            if key in self._configured_checks:
                                yaml_config = self._configured_checks[key]
                                effective_config = dict_merge(
                                    effective_config, yaml_config, in_place=True
                                )

                                # Make sure there's nothing persisted in that check that shouldn't be persisted
                                #
                                # This used to be part of the settings migration (version 2) due to a bug - it can't
                                # stay there though since it interferes with manual entries to the checks not
                                # originating from within a plugin. Hence we do that step now here.
                                if (
                                    "type" not in effective_config
                                    or effective_config["type"]
                                    not in self.CURRENT_TRACKING_TYPES
                                ):
                                    deletables = ["current", "displayVersion"]
                                else:
                                    deletables = []
                                self._clean_settings_check(
                                    key,
                                    yaml_config,
                                    default_config,
                                    delete=deletables,
                                    save=False,
                                )

                            if effective_config:
                                effective_configs[key] = effective_config
                            else:
                                self._logger.warning(
                                    "Update for {} is empty or None, ignoring it".format(
                                        key
                                    )
                                )

                # finally set all our internal representations to our processed results
                for key, config in effective_configs.items():
                    self._configured_checks[key] = config

                # we only want to process checks that came from plugins for
                # which the plugins are still installed and enabled
                config_checks = self._settings.get(["checks"])
                plugin_and_not_enabled = (
                    lambda k: k in check_providers
                    and check_providers[k] not in self._plugin_manager.enabled_plugins
                )
                obsolete_plugin_checks = list(
                    filter(plugin_and_not_enabled, config_checks.keys())
                )
                for key in obsolete_plugin_checks:
                    self._logger.debug(
                        "Check for key {} was provided by plugin {} that's no longer available, ignoring it".format(
                            key, check_providers[key]
                        )
                    )
                    del self._configured_checks[key]

            return self._configured_checks

    def _check_environment(self):
        import pkg_resources

        local_pip = create_pip_caller(
            command=self._settings.global_get(["server", "commands", "localPipCommand"])
        )

        # check python and setuptools version
        versions = {
            "python": get_python_version_string(),
            "setuptools": pkg_resources.get_distribution("setuptools").version,
            "pip": local_pip.version_string,
        }
        supported = (
            get_comparable_version(versions["python"])
            >= get_comparable_version(MINIMUM_PYTHON)
            and get_comparable_version(versions["setuptools"])
            >= get_comparable_version(MINIMUM_SETUPTOOLS)
            and get_comparable_version(versions["pip"])
            >= get_comparable_version(MINIMUM_PIP)
        )

        self._environment_supported = supported
        self._environment_versions = versions
        self._environment_ready.set()

    def _check_storage(self):
        import distutils.sysconfig
        import tempfile

        import psutil

        storage_info = {}
        paths = {
            "python": distutils.sysconfig.get_python_lib(),
            "plugins": self._settings.global_get_basefolder("plugins"),
            "temp": tempfile.gettempdir(),
        }

        for key, path in paths.items():
            info = {"path": path, "free": None}

            try:
                data = psutil.disk_usage(path)
                info["free"] = data.free
            except Exception:
                self._logger.exception(f"Error while determining disk usage of {path}")
                continue

            storage_info[key] = info

        if len(storage_info):
            free_storage = min(*list(map(lambda x: x["free"], storage_info.values())))
        else:
            free_storage = None

        self._storage_sufficient = (
            free_storage is None
            or free_storage
            >= self._settings.get_int(["minimum_free_storage"]) * 1024 * 1024
        )
        self._storage_info = storage_info

        self._logger.info(
            "Minimum free storage across all update relevant locations is {}. "
            "That is considered {} for updating.".format(
                get_formatted_size(free_storage)
                if free_storage is not None
                else "unknown",
                "sufficient" if self._storage_sufficient else "insufficient",
            )
        )

    @property
    def _check_overlays_stale(self):
        return (
            self._overlay_cache is None
            or self._overlay_cache_timestamp is None
            or self._overlay_cache_timestamp + self._overlay_cache_ttl < time.time()
        )

    def _get_check_overlay(self, url):
        self._logger.info(f"Fetching check overlays from {url}")
        try:
            r = requests.get(url, timeout=3.1)
            r.raise_for_status()
            data = r.json()
        except Exception as exc:
            self._logger.error(f"Could not fetch check overlay from {url}: {exc}")
            return {}
        else:
            return data

    def _get_check_overlays(self, force=False):
        if self._check_overlays_stale or force:
            if self._connectivity_checker.online:
                self._overlay_cache = self._get_check_overlay(
                    self._settings.get(["check_overlay_url"])
                )
                if is_python_compatible("<3"):
                    data = self._get_check_overlay(
                        self._settings.get(["check_overlay_py2_url"])
                    )
                    self._overlay_cache = octoprint.util.dict_merge(
                        self._overlay_cache, data
                    )

            else:
                self._logger.info("Not fetching check overlays, we are offline")
                self._overlay_cache = {}

            self._overlay_cache_timestamp = time.time()

            default_overlay = {}
            defaults = self.get_settings_defaults()
            for key in defaults["checks"]:
                if key in self._overlay_cache:
                    default_overlay[key] = self._overlay_cache[key]

            self._settings.remove_overlay(self.CHECK_OVERLAY_KEY)
            if default_overlay:
                self._settings.add_overlay(
                    {"checks": default_overlay}, key=self.CHECK_OVERLAY_KEY
                )

        return self._overlay_cache

    def _load_version_cache(self):
        if not os.path.isfile(self._version_cache_path):
            return

        try:
            data = yaml.load_from_file(path=self._version_cache_path)
            timestamp = os.stat(self._version_cache_path).st_mtime
        except Exception:
            self._logger.exception("Error while loading version cache from disk")
        else:
            try:
                if not isinstance(data, dict):
                    self._logger.info(
                        "Version cache was created in a different format, not using it"
                    )
                    return

                if "__version" in data:
                    data_version = data["__version"]
                else:
                    self._logger.info(
                        "Can't determine version of OctoPrint version cache was created for, not using it"
                    )
                    return

                from octoprint._version import get_versions

                octoprint_version = get_versions()["version"]
                if data_version != octoprint_version:
                    self._logger.info(
                        "Version cache was created for another version of OctoPrint, not using it"
                    )
                    return

                self._version_cache = data
                self._version_cache_dirty = False
                self._version_cache_timestamp = timestamp
                self._logger.info("Loaded version cache from disk")
            except Exception:
                self._logger.exception("Error parsing in version cache data")

    def _save_version_cache(self):
        from octoprint._version import get_versions
        from octoprint.util import atomic_write

        octoprint_version = get_versions()["version"]
        self._version_cache["__version"] = octoprint_version

        with atomic_write(
            self._version_cache_path, mode="wt", max_permissions=0o666
        ) as file_obj:
            yaml.save_to_file(self._version_cache, file=file_obj, pretty=True)

        self._version_cache_dirty = False
        self._version_cache_timestamp = time.time()
        self._logger.info("Saved version cache to disk")

    def _invalidate_version_cache(self, target):
        self._refresh_configured_checks = True
        try:
            del self._version_cache[target]
        except KeyError:
            pass
        self._version_cache_dirty = True

    def _load_update_log(self):
        if not os.path.isfile(self._update_log_path):
            return

        try:
            data = yaml.load_from_file(path=self._update_log_path)
        except Exception:
            self._logger.exception("Error while loading update log from disk")
        else:
            # clean up data older than updatelog_cutoff hours
            before_cleanup = len(data)
            cutoff = (
                datetime.datetime.now()
                - datetime.timedelta(minutes=self._settings.get_int(["updatelog_cutoff"]))
            ).strftime(DATETIME_FORMAT)
            data = sorted(
                list(
                    filter(
                        lambda x: x.get("datetime") and x["datetime"] > cutoff,
                        data,
                    )
                ),
                key=lambda x: x["datetime"],
            )
            cleaned_up = len(data) - before_cleanup
            if cleaned_up:
                self._logger.info(f"Cleaned up {cleaned_up} old update log entries")

            with self._update_log_mutex:
                self._update_log = data
                self._update_log_dirty = False

            self._logger.info("Loaded update log from disk")

    def _save_update_log(self):
        from octoprint.util import atomic_write

        with self._update_log_mutex:
            with atomic_write(
                self._update_log_path, mode="wt", max_permissions=0o666
            ) as file_obj:
                yaml.save_to_file(
                    sorted(self._update_log, key=lambda x: x["datetime"]),
                    file=file_obj,
                    pretty=True,
                )
                self._update_log_dirty = False

        self._logger.info("Saved update log to disk")

    def _add_update_log_entry(
        self,
        target,
        current_version,
        target_version,
        success,
        text,
        display_name=None,
        release_notes="",
        save=True,
    ):
        if display_name is None:
            display_name = target

        with self._update_log_mutex:
            self._update_log.append(
                {
                    "datetime": datetime.datetime.now().strftime(DATETIME_FORMAT),
                    "target": target,
                    "current_version": current_version,
                    "target_version": target_version,
                    "display_name": display_name,
                    "release_notes": release_notes if release_notes else "",
                    "success": success,
                    "text": text,
                },
            )
            self._update_log_dirty = True

            if save:
                self._save_update_log()

    def _is_octoprint_outdated_and_can_update(self):
        checks = self._get_configured_checks()
        check = checks.get("octoprint", None)
        if not check:
            return False

        check = self._populated_check("octoprint", check)
        credentials = self._settings.get(["credentials"], merged=True)

        _, update_available, update_possible, _, _ = self._get_current_version(
            "octoprint", check, credentials=credentials
        )
        if not update_available or not update_possible:
            return False

        restart_type = self._get_restart_type(check)
        return self._has_restart_command(restart_type)

    def _get_restart_type(self, check):
        if check.get("restart") in self.VALID_RESTART_TYPES:
            return check["restart"]
        elif "pip" in check or check.get("method") in self.OCTOPRINT_RESTART_TYPES:
            return "octoprint"
        else:
            target_restart_type = None

        return target_restart_type

    def _has_restart_command(self, restart_type):
        if restart_type == "octoprint":
            return self._system_commands.has_server_restart_command()
        elif restart_type == "environment":
            return self._system_commands.has_system_restart_command()
        else:
            return False

    # ~~ SettingsPlugin API

    def get_settings_defaults(self):
        update_script = os.path.join(self._basefolder, "scripts", "update-octoprint.py")
        default_update_script = (
            '{{python}} "{update_script}" --branch={{branch}} '
            '--force={{force}} "{{folder}}" {{target}}'.format(
                update_script=update_script
            )
        )

        return {
            "checks": {
                "octoprint": {
                    "type": "github_release",
                    "user": "foosel",
                    "repo": "OctoPrint",
                    "method": "pip",
                    "pip": "https://github.com/OctoPrint/OctoPrint/archive/{target_version}.zip",
                    "update_script": default_update_script,
                    "restart": "octoprint",
                    "stable_branch": {
                        "branch": "master",
                        "commitish": ["master"],
                        "name": "Stable",
                    },
                    "prerelease_branches": [
                        {
                            "branch": "rc/maintenance",
                            "commitish": ["rc/maintenance"],  # maintenance RCs
                            "name": "Maintenance RCs",
                        },
                        {
                            "branch": "rc/devel",
                            "commitish": [
                                "rc/maintenance",
                                "rc/devel",
                            ],  # devel & maintenance RCs
                            "name": "Devel RCs",
                        },
                    ],
                },
            },
            "pip_command": None,
            "cache_ttl": 24 * 60,  # 1 day
            "notify_users": True,
            "ignore_throttled": False,
            "minimum_free_storage": 150,
            "check_overlay_url": "https://plugins.octoprint.org/update_check_overlay.json",
            "check_overlay_py2_url": "https://plugins.octoprint.org/update_check_overlay_py2.json",
            "check_overlay_ttl": 6 * 60,  # 6 hours
            "updatelog_cutoff": 30 * 24 * 60,  # 30 days
            "credentials": {
                "github": None,
                "bitbucket_user": None,
                "bitbucket_password": None,
            },
        }

    def on_settings_load(self):
        # ensure we don't persist check configs we receive on the API
        data = dict(octoprint.plugin.SettingsPlugin.on_settings_load(self))
        if "checks" in data:
            del data["checks"]

        checks = self._get_configured_checks()

        # ~~ octoprint check settings

        if "octoprint" in checks:
            data["octoprint_checkout_folder"] = self._get_octoprint_checkout_folder(
                checks=checks
            )
            data["octoprint_tracked_branch"] = self._get_octoprint_tracked_branch(
                checks=checks
            )
            data["octoprint_pip_target"] = self._get_octoprint_pip_target(checks=checks)
            data["octoprint_type"] = checks["octoprint"].get("type", None)

            try:
                data["octoprint_method"] = self._get_update_method(
                    "octoprint", checks["octoprint"]
                )
            except exceptions.UnknownUpdateType:
                data["octoprint_method"] = "unknown"

            stable_branch = None
            prerelease_branches = []
            branch_mappings = []
            if "stable_branch" in checks["octoprint"]:
                branch_mappings.append(checks["octoprint"]["stable_branch"])
                stable_branch = checks["octoprint"]["stable_branch"]["branch"]
            if "prerelease_branches" in checks["octoprint"]:
                for mapping in checks["octoprint"]["prerelease_branches"]:
                    branch_mappings.append(mapping)
                    prerelease_branches.append(mapping["branch"])
            data["octoprint_branch_mappings"] = branch_mappings

            data["octoprint_release_channel"] = stable_branch
            if checks["octoprint"].get("prerelease", False):
                channel = checks["octoprint"].get("prerelease_channel", BRANCH)
                if channel in prerelease_branches:
                    data["octoprint_release_channel"] = channel

        else:
            data["octoprint_checkout_folder"] = None
            data["octoprint_type"] = None
            data["octoprint_release_channel"] = None

        # ~~ pip check settings

        data["pip_enable_check"] = "pip" in checks

        data["queued_updates"] = self._queued_updates.get("targets", [])

        return data

    def get_settings_restricted_paths(self):
        return {"never": [["credentials"]]}

    def on_settings_save(self, data):
        # ~~ plugin settings

        for key in self.get_settings_defaults():
            if key in (
                "checks",
                "cache_ttl",
                "check_overlay_ttl",
                "notify_user",
                "minimum_free_storage",
                "octoprint_checkout_folder",
                "octoprint_type",
                "octoprint_release_channel",
            ):
                continue
            if key in data:
                self._settings.set([key], data[key])

        if "cache_ttl" in data:
            self._settings.set_int(["cache_ttl"], data["cache_ttl"])
        self._version_cache_ttl = self._settings.get_int(["cache_ttl"]) * 60

        if "check_overlay_ttl" in data:
            self._settings.set_int(["check_overlay_ttl"], data["check_overlay_ttl"])
        self._overlay_cache_ttl = self._settings.get_int(["check_overlay_ttl"]) * 60

        if "notify_users" in data:
            self._settings.set_boolean(["notify_users"], data["notify_users"])

        if "minimum_free_storage" in data:
            self._settings.set_int(["minimum_free_storage"], data["minimum_free_storage"])
            self._check_storage()

        # ~~ octoprint check settings

        updated_octoprint_check_config = False

        if "octoprint_checkout_folder" in data:
            if data["octoprint_checkout_folder"]:
                # force=True as that path is not part of the default config
                self._settings.set(
                    ["checks", "octoprint", "checkout_folder"],
                    data["octoprint_checkout_folder"],
                    force=True,
                )
                if self._settings.get(["checks", "octoprint", "update_folder"]):
                    self._settings.remove(["checks", "octoprint", "update_folder"])
            else:
                self._settings.remove(["checks", "octoprint", "checkout_folder"])
            updated_octoprint_check_config = True

        if "octoprint_type" in data:
            octoprint_type = data["octoprint_type"]

            if octoprint_type == "github_release":
                self._settings.set(["checks", "octoprint", "type"], octoprint_type)
                self._settings.set(["checks", "octoprint", "method"], "pip")
                updated_octoprint_check_config = True

            elif octoprint_type == "github_commit":
                self._settings.set(["checks", "octoprint", "type"], octoprint_type)
                self._settings.set(["checks", "octoprint", "method"], "pip")
                updated_octoprint_check_config = True

            elif octoprint_type == "git_commit":
                self._settings.set(["checks", "octoprint", "type"], octoprint_type)
                self._settings.set(["checks", "octoprint", "method"], "update_script")
                updated_octoprint_check_config = True

        if "octoprint_tracked_branch" in data:
            if data["octoprint_tracked_branch"]:
                # force=True as that path is not part of the default config
                self._settings.set(
                    ["checks", "octoprint", "branch"],
                    data["octoprint_tracked_branch"],
                    force=True,
                )
            else:
                self._settings.remove(["checks", "octoprint", "branch"])
            updated_octoprint_check_config = True

        if "octoprint_pip_target" in data:
            self._settings.set(
                ["checks", "octoprint", "pip"], data["octoprint_pip_target"]
            )
            updated_octoprint_check_config = True

        if "octoprint_release_channel" in data:
            prerelease_branches = self._settings.get(
                ["checks", "octoprint", "prerelease_branches"]
            )
            if prerelease_branches and data["octoprint_release_channel"] in [
                x["branch"] for x in prerelease_branches
            ]:
                # force=True as those paths are not part of the default config
                self._settings.set(
                    ["checks", "octoprint", "prerelease"], True, force=True
                )
                self._settings.set(
                    ["checks", "octoprint", "prerelease_channel"],
                    data["octoprint_release_channel"],
                    force=True,
                )
            else:
                self._settings.remove(["checks", "octoprint", "prerelease"])
                self._settings.remove(["checks", "octoprint", "prerelease_channel"])
            updated_octoprint_check_config = True

        if updated_octoprint_check_config:
            self._invalidate_version_cache("octoprint")

        # ~~ pip check settings

        update_pip_check_config = False

        if "pip_enable_check" in data:
            pip_enable_check = (
                data["pip_enable_check"] in octoprint.settings.valid_boolean_trues
            )
            pip_check_currently_enabled = "pip" in self._settings.get(
                ["checks"], merged=True
            )

            if pip_enable_check != pip_check_currently_enabled:
                if pip_enable_check:
                    # force=True as that path is not part of the default config
                    self._settings.set(
                        ["checks", "pip"],
                        {
                            "type": "pypi_release",
                            "package": "pip",
                            "pip": "pip=={target_version}",
                        },
                        force=True,
                    )
                else:
                    self._settings.remove(["checks", "pip"])
                update_pip_check_config = True

        if update_pip_check_config:
            self._invalidate_version_cache("pip")

    def get_settings_version(self):
        return 9

    def on_settings_migrate(self, target, current=None):

        if current is None or current < 6:
            # up until & including config version 5 we didn't set the method parameter for the octoprint check
            # configuration

            configured_checks = self._settings.get(["checks"], incl_defaults=False)
            if configured_checks is not None and "octoprint" in configured_checks:
                octoprint_check = dict(configured_checks["octoprint"])

                if (
                    "method" not in octoprint_check
                    and octoprint_check.get("type") == "git_commit"
                ):
                    defaults = {
                        "plugins": {
                            "softwareupdate": {"checks": {"octoprint": {"method": "pip"}}}
                        }
                    }
                    self._settings.set(
                        ["checks", "octoprint", "method"],
                        "update_script",
                        defaults=defaults,
                    )

        if current == 4:
            # config version 4 didn't correctly remove the old settings for octoprint_restart_command
            # and environment_restart_command

            self._settings.set(["environment_restart_command"], None)
            self._settings.set(["octoprint_restart_command"], None)

        if current is None or current < 5:
            # config version 4 and higher moves octoprint_restart_command and
            # environment_restart_command to the core configuration

            # current plugin commands
            configured_octoprint_restart_command = self._settings.get(
                ["octoprint_restart_command"]
            )
            configured_environment_restart_command = self._settings.get(
                ["environment_restart_command"]
            )

            # current global commands
            configured_system_restart_command = self._settings.global_get(
                ["server", "commands", "systemRestartCommand"]
            )
            configured_server_restart_command = self._settings.global_get(
                ["server", "commands", "serverRestartCommand"]
            )

            # only set global commands if they are not yet set
            if (
                configured_system_restart_command is None
                and configured_environment_restart_command is not None
            ):
                self._settings.global_set(
                    ["server", "commands", "systemRestartCommand"],
                    configured_environment_restart_command,
                )
            if (
                configured_server_restart_command is None
                and configured_octoprint_restart_command is not None
            ):
                self._settings.global_set(
                    ["server", "commands", "serverRestartCommand"],
                    configured_octoprint_restart_command,
                )

            # delete current plugin commands from config
            self._settings.set(["environment_restart_command"], None)
            self._settings.set(["octoprint_restart_command"], None)

        if current is None or current == 2 or current == 8:
            # No config version and config version 2 and 8 need the same fix, stripping
            # accidentally persisted data off the checks.
            #
            # We used to do the same processing for the plugin entries too here, but that interfered
            # with manual configuration entries. Stuff got deleted that wasn't supposed to be deleted.
            #
            # The problem is that we don't know if an entry we are looking at and which didn't come through
            # a plugin hook is simply an entry from a now uninstalled/unactive plugin, or if it was something
            # manually configured by the user. So instead of just blindly removing anything that doesn't
            # come from a plugin here we instead clean up anything that indeed comes from a plugin
            # during run time and leave everything else as is in the hopes that will not cause trouble.
            #
            # We still handle the "octoprint" entry here though.

            configured_checks = self._settings.get(["checks"], incl_defaults=False)
            if configured_checks is not None and "octoprint" in configured_checks:
                octoprint_check = dict(configured_checks["octoprint"])
                if (
                    "type" not in octoprint_check
                    or octoprint_check["type"] not in self.CURRENT_TRACKING_TYPES
                ):
                    deletables = ["current", "displayName", "displayVersion"]
                else:
                    deletables = []
                self._clean_settings_check(
                    "octoprint",
                    octoprint_check,
                    self.get_settings_defaults()["checks"]["octoprint"],
                    delete=deletables,
                    save=False,
                )

        elif current == 1:
            # config version 1 had the error that the octoprint check got accidentally
            # included in checks["octoprint"], leading to recursion and hence to
            # yaml parser errors

            configured_checks = self._settings.get(["checks"], incl_defaults=False)
            if configured_checks is None:
                return

            if (
                "octoprint" in configured_checks
                and "octoprint" in configured_checks["octoprint"]
            ):
                # that's a circular reference, back to defaults
                dummy_defaults = {"plugins": {}}
                dummy_defaults["plugins"][self._identifier] = {"checks": {}}
                dummy_defaults["plugins"][self._identifier]["checks"]["octoprint"] = None
                self._settings.set(["checks", "octoprint"], None, defaults=dummy_defaults)

        if current is None or current < 8:
            # remove check_providers again
            self._settings.remove(["check_providers"])

    def _clean_settings_check(self, key, data, defaults, delete=None, save=True):
        if not data:
            # nothing to do
            return data

        if delete is None:
            delete = []

        for k in data:
            if k in defaults and defaults[k] == data[k]:
                delete.append(k)

        for k in delete:
            if k in data:
                del data[k]

        dummy_defaults = {"plugins": {}}
        dummy_defaults["plugins"][self._identifier] = {"checks": {}}
        dummy_defaults["plugins"][self._identifier]["checks"][key] = defaults
        if len(data):
            self._settings.set(["checks", key], data, defaults=dummy_defaults)
        else:
            self._settings.remove(["checks", key], defaults=dummy_defaults)

        if save:
            self._settings.save()

        return data

    # ~~ BluePrint API

    @octoprint.plugin.BlueprintPlugin.route("/updatelog", methods=["GET"])
    @no_firstrun_access
    @Permissions.ADMIN.require(403)
    def get_update_log(self):
        def view():
            return flask.jsonify(updatelog=list(reversed(self._update_log)))

        def etag(lm=None):
            hash = hashlib.sha1()

            def hash_update(value):
                value = value.encode("utf-8")
                hash.update(value)

            hash_update(repr(self._update_log))
            return hash.hexdigest()

        def lastmodified():
            if os.path.exists(self._update_log_path):
                return os.stat(self._update_log_path).st_mtime
            else:
                return 0

        return with_revalidation_checking(
            etag_factory=etag, lastmodified_factory=lastmodified
        )(view)()

    @octoprint.plugin.BlueprintPlugin.route("/check", methods=["GET"])
    @no_firstrun_access
    @Permissions.PLUGIN_SOFTWAREUPDATE_CHECK.require(403)
    def check_for_update(self):
        if (
            not Permissions.PLUGIN_SOFTWAREUPDATE_CHECK.can()
            and not self._settings.global_get(["server", "firstRun"])
        ):
            flask.abort(403)

        request_data = flask.request.values

        if "targets" in request_data or "check" in request_data:
            check_targets = list(
                map(
                    lambda x: x.strip(),
                    request_data.get("targets", request_data.get("check", "")).split(","),
                )
            )
        else:
            check_targets = None

        force = (
            request_data.get("force", "false") in octoprint.settings.valid_boolean_trues
        )

        def view():
            self._environment_ready.wait(timeout=30.0)

            try:
                (
                    information,
                    update_available,
                    update_possible,
                ) = self.get_current_versions(check_targets=check_targets, force=force)

                storage = list()
                for key, name in (
                    ("python", gettext("Python package installation folder")),
                    ("plugins", gettext("Plugin folder")),
                    ("temp", gettext("System temporary files")),
                ):
                    data = self._storage_info.get(key)
                    if not data:
                        continue

                    s = {"name": name}
                    s.update(**data)
                    storage.append(s)

                status = "current"
                if self._update_in_progress:
                    status = "inProgress"
                elif (
                    update_available
                    and update_possible
                    and self._environment_supported
                    and self._storage_sufficient
                ):
                    status = "updatePossible"
                elif update_available:
                    status = "updateAvailable"

                return flask.jsonify(
                    {
                        "status": status,
                        "information": information,
                        "timestamp": self._version_cache_timestamp,
                        "environment": {
                            "supported": self._environment_supported,
                            "versions": [
                                {
                                    "name": gettext("Python"),
                                    "current": self._environment_versions.get(
                                        "python", "unknown"
                                    ),
                                    "minimum": MINIMUM_PYTHON,
                                },
                                {
                                    "name": gettext("pip"),
                                    "current": self._environment_versions.get(
                                        "pip", "unknown"
                                    ),
                                    "minimum": MINIMUM_PIP,
                                },
                                {
                                    "name": gettext("setuptools"),
                                    "current": self._environment_versions.get(
                                        "setuptools", "unknown"
                                    ),
                                    "minimum": MINIMUM_SETUPTOOLS,
                                },
                            ],
                        },
                        "storage": {
                            "sufficient": self._storage_sufficient,
                            "free": storage,
                        },
                    }
                )
            except exceptions.ConfigurationInvalid as e:
                flask.abort(
                    500,
                    description="Update not properly configured, can't proceed: {}".format(
                        e
                    ),
                )

        def etag():
            checks = self._get_configured_checks()

            targets = check_targets
            if targets is None:
                targets = checks.keys()

            hash = hashlib.sha1()

            def hash_update(value):
                value = value.encode("utf-8")
                hash.update(value)

            targets = sorted(targets)
            for target in targets:
                current_hash = self._get_check_hash(checks.get(target, {}))
                if target in self._version_cache and not force:
                    data = self._version_cache[target]
                    hash_update(current_hash)
                    hash_update(
                        str(
                            data["timestamp"] + self._version_cache_ttl
                            >= time.time()
                            > data["timestamp"]
                        )
                    )
                    hash_update(repr(data["information"]))
                    hash_update(str(data["available"]))
                    hash_update(str(data["possible"]))
                    hash_update(str(data.get("online", None)))

            hash_update(",".join(targets))

            hash_update(str(self._environment_supported))
            hash_update(str(self._version_cache_timestamp))
            hash_update(str(self._overlay_cache_timestamp))
            hash_update(str(self._connectivity_checker.online))
            hash_update(str(self._update_in_progress))
            hash_update(self.DATA_FORMAT_VERSION)
            return hash.hexdigest()

        def condition():
            return check_etag(etag())

        return with_revalidation_checking(
            etag_factory=lambda *args, **kwargs: etag(),
            condition=lambda *args, **kwargs: condition(),
            unless=lambda: force or self._check_overlays_stale,
        )(view)()

    @octoprint.plugin.BlueprintPlugin.route("/update", methods=["POST"])
    def perform_update(self):
        if (
            not Permissions.PLUGIN_SOFTWAREUPDATE_UPDATE.can()
            and not self._settings.global_get(["server", "firstRun"])
        ):
            flask.abort(403)

        throttled = self._get_throttled()
        if (
            throttled
            and isinstance(throttled, dict)
            and throttled.get("current_issue", False)
            and not self._settings.get_boolean(["ignore_throttled"])
        ):
            # currently throttled, we refuse to run
            message = (
                "System is currently throttled, refusing to update "
                "anything due to possible stability issues"
            )
            self._logger.error(message)
            flask.abort(
                409,
                description=message,
            )

        if not self._environment_supported:
            flask.abort(
                409,
                description="Direct updates are not supported in this Python environment",
            )

        if not self._storage_sufficient:
            flask.abort(409, description="Not enough free disk space for updating")

        if "application/json" not in flask.request.headers["Content-Type"]:
            flask.abort(400, description="Expected content-type JSON")

        json_data = flask.request.get_json(silent=True)
        if json_data is None:
            flask.abort(400, description="Invalid JSON")

        if "targets" in json_data or "checks" in json_data:
            targets = list(
                map(
                    lambda x: x.strip(),
                    json_data.get("targets", json_data.get("check", [])),
                )
            )
        else:
            targets = None

        force = json_data.get("force", "false") in octoprint.settings.valid_boolean_trues

        if self._printer.is_printing() or self._printer.is_paused():
            # do not update while a print job is running
            # store targets to be run later on print done event
            self._queued_updates["targets"] = list(
                set(self._queued_updates["targets"] + targets)
            )
            self._send_client_message(
                "queued_updates",
                {"targets": self._queued_updates["targets"]},
            )

            return (
                flask.jsonify({"queued": self._queued_updates.get("targets", False)}),
                202,
            )

        to_be_checked, checks = self.perform_updates(targets=targets, force=force)
        return flask.jsonify({"order": to_be_checked, "checks": checks})

    @octoprint.plugin.BlueprintPlugin.route("/update/queued", methods=["POST"])
    @no_firstrun_access
    @Permissions.PLUGIN_SOFTWAREUPDATE_UPDATE.require(403)
    def cancel_queued(self):
        if "application/json" not in flask.request.headers["Content-Type"]:
            flask.abort(400, description="Expected content-type JSON")

        json_data = flask.request.get_json(silent=True)
        if json_data is None:
            flask.abort(400, description="Invalid JSON")

        command = json_data.get("command")
        if command == "cancel":
            targets = [x.strip() for x in json_data.get("targets", [])]

            if targets:
                self._queued_updates["targets"] = [
                    x for x in self._queued_updates["targets"] if x not in targets
                ]
            else:
                self._queued_updates["targets"] = []

            if not self._queued_updates["targets"] and self._queued_updates_abort_timer:
                self._queued_updates_abort_timer.cancel()
                self._queued_updates_abort_timer = None

            self._send_client_message(
                "queued_updates",
                {"targets": self._queued_updates["targets"]},
            )

            return (
                flask.jsonify({"queued": self._queued_updates["targets"]}),
                202,
            )

        else:
            flask.abort(400, description="Expected a valid command")

    @octoprint.plugin.BlueprintPlugin.route("/configure", methods=["POST"])
    @no_firstrun_access
    @Permissions.PLUGIN_SOFTWAREUPDATE_CONFIGURE.require(403)
    def configure_update(self):
        json_data = flask.request.get_json(silent=True)
        if json_data is None:
            flask.abort(400)

        checks = self._get_configured_checks()

        settings_dirty = False
        for target, data in json_data.items():
            if target not in checks:
                continue

            try:
                populated_check = self._populated_check(target, checks[target])
            except exceptions.UnknownCheckType:
                self._logger.debug(f"Ignoring unknown check type for target {target}")
                continue
            except Exception:
                self._logger.exception(
                    "Error while populating check prior to configuration for target {}".format(
                        target
                    )
                )
                continue

            patch = {}

            # switched release channel
            if "channel" in data:
                if (
                    populated_check["type"] == "github_release"
                    and "stable_branch" in populated_check
                    and "prerelease_branches" in populated_check
                ):
                    valid_channel = data["channel"] == populated_check[
                        "stable_branch"
                    ].get("branch") or any(
                        map(
                            lambda x: data["channel"] == x.get("branch"),
                            populated_check["prerelease_branches"],
                        )
                    )
                    prerelease = data["channel"] != populated_check["stable_branch"].get(
                        "branch"
                    )

                    if valid_channel:
                        patch["prerelease"] = prerelease
                        patch["prerelease_channel"] = data["channel"]

            # disable/enable check
            if "disabled" in data:
                patch["disabled"] = (
                    data["disabled"] in octoprint.settings.valid_boolean_trues
                )

            # do we have changes to apply?
            if patch:
                current = self._settings.get(["checks", target], incl_defaults=False)
                updated = dict_merge(current, patch)
                self._settings.set(
                    ["checks", target],
                    updated,
                    defaults={"plugins": {"softwareupdate": {"checks": {target: {}}}}},
                )

                settings_dirty = True
                self._invalidate_version_cache(target)

        if settings_dirty:
            self._settings.save()

        return NO_CONTENT

    def is_blueprint_protected(self):
        return False

    def is_blueprint_csrf_protected(self):
        return True

    # ~~ Asset API

    def get_assets(self):
        return {
            "css": ["css/softwareupdate.css"],
            "js": ["js/softwareupdate.js"],
            "clientjs": ["clientjs/softwareupdate.js"],
            "less": ["less/softwareupdate.less"],
        }

    ##~~ TemplatePlugin API

    def get_template_configs(self):
        from flask_babel import gettext

        templates = [
            {"type": "settings", "name": gettext("Software Update")},
        ]

        if self._is_wizard_update_required():
            templates.append(
                {
                    "type": "wizard",
                    "name": gettext("Update?"),
                    "template": "softwareupdate_wizard_update.jinja2",
                    "suffix": "_update",
                }
            )

        if self._is_settings_wizard_required():
            templates.append(
                {
                    "type": "wizard",
                    "name": gettext("Software Update"),
                    "template": "softwareupdate_wizard_settings.jinja2",
                    "suffix": "_settings",
                }
            )

        return templates

    ##~~ WizardPlugin API

    def _is_wizard_update_required(self):
        firstrun = self._settings.global_get(["server", "firstRun"])
        return (
            firstrun
            and self._is_octoprint_outdated_and_can_update()
            and not self._is_settings_wizard_required()
        )

    def _is_settings_wizard_required(self):
        checks = self._get_configured_checks()
        check = checks.get("octoprint", None)
        checkout_folder = self._get_octoprint_checkout_folder(checks=checks)
        return check and check.get("method") == "update_script" and not checkout_folder

    def is_wizard_required(self):
        return self._is_wizard_update_required() or self._is_settings_wizard_required()

    def get_wizard_version(self):
        return 1

    def get_wizard_details(self):
        result = {}
        if self._is_wizard_update_required():
            information, _, _ = self.get_current_versions()
            data = information.get("octoprint", {})
            result["update"] = data
        return result

    ##~~ EventHandlerPlugin API

    def on_event(self, event, payload):
        from octoprint.events import Events

        if event == Events.PRINT_STARTED:
            self._queued_updates_timer_stop()
            self._print_cancelled = False
        elif (
            event == Events.PRINT_DONE
            and self._settings.global_get(["webcam", "timelapse", "type"]) == "off"
            and len(self._queued_updates.get("targets", [])) > 0
        ):
            self._queued_updates_timer_start()
        elif (
            event == Events.PRINT_FAILED
            and len(self._queued_updates.get("targets", [])) > 0
        ):
            self._send_client_message(
                "queued_updates",
                {"print_failed": True, "targets": self._queued_updates["targets"]},
            )
            self._print_cancelled = True
        elif (
            event == Events.MOVIE_DONE
            and self._settings.global_get(["webcam", "timelapse", "type"]) != "off"
            and len(self._queued_updates.get("targets", [])) > 0
            and not (self._printer.is_printing() or self._printer.is_paused())
            and not self._print_cancelled
        ):
            self._queued_updates_timer_start()
        elif (
            event != Events.CONNECTIVITY_CHANGED
            or not payload
            or not payload.get("new", False)
        ):
            return

        thread = threading.Thread(target=self.get_current_versions)
        thread.daemon = True
        thread.start()

    # ~~ Updater

    def get_current_versions(self, check_targets=None, force=False):
        """
        Retrieves the current version information for all defined check_targets. Will retrieve information for all
        available targets by default.

        :param check_targets: an iterable defining the targets to check, if not supplied defaults to all targets
        """

        if force:
            self._get_check_overlays(force=True)

        checks = self._get_configured_checks()
        if check_targets is None:
            check_targets = list(checks.keys())

        # TODO: caching doesn't take check_targets into account!

        update_available = False
        update_possible = False
        information = {}

        credentials = self._settings.get(["credentials"], merged=True)

        # we don't want to do the same work twice, so let's use a lock
        if self._get_versions_mutex.acquire(False):
            self._get_versions_data_ready.clear()
            try:
                futures_to_result = {}
                online = self._connectivity_checker.check_immediately()
                self._logger.debug(
                    "Looks like we are {}".format("online" if online else "offline")
                )

                with futures.ThreadPoolExecutor(max_workers=5) as executor:
                    for target, check in checks.items():
                        if target not in check_targets:
                            continue

                        if not check:
                            continue

                        if "type" not in check:
                            continue

                        try:
                            populated_check = self._populated_check(target, check)
                            future = executor.submit(
                                self._get_current_version,
                                target,
                                populated_check,
                                force=force,
                                credentials=credentials,
                            )
                            futures_to_result[future] = (target, populated_check)
                        except exceptions.UnknownCheckType:
                            self._logger.warning(
                                "Unknown update check type for target {}: {}".format(
                                    target, check.get("type", "<n/a>")
                                )
                            )
                            continue
                        except Exception:
                            self._logger.exception(
                                f"Could not check {target} for updates"
                            )
                            continue

                    for future in futures.as_completed(futures_to_result):

                        target, populated_check = futures_to_result[future]
                        if future.exception() is not None:
                            self._logger.error(
                                "Could not check {} for updates, error: {!r}".format(
                                    target, future.exception()
                                )
                            )
                            continue

                        (
                            target_information,
                            target_update_available,
                            target_update_possible,
                            target_online,
                            target_error,
                        ) = future.result()
                        target_update_possible = (
                            target_update_possible and self._environment_supported
                        )

                        target_information = dict_merge(
                            {
                                "local": {"name": "?", "value": "?"},
                                "remote": {
                                    "name": "?",
                                    "value": "?",
                                    "release_notes": None,
                                },
                                "needs_online": True,
                            },
                            target_information,
                        )

                        target_disabled = populated_check.get("disabled", False)
                        target_compatible = populated_check.get("compatible", True)
                        update_available = update_available or (
                            target_update_available and not target_disabled
                        )
                        update_possible = update_possible or (
                            target_update_possible
                            and target_update_available
                            and not target_disabled
                        )

                        local_name = target_information["local"]["name"]
                        local_value = target_information["local"]["value"]

                        release_notes = self._get_release_notes(
                            target_information, populated_check
                        )

                        information[target] = {
                            "updateAvailable": target_update_available,
                            "updatePossible": target_update_possible,
                            "information": target_information,
                            "displayName": populated_check["displayName"],
                            "displayVersion": populated_check["displayVersion"].format(
                                octoprint_version=VERSION,
                                local_name=local_name,
                                local_value=local_value,
                            ),
                            "releaseNotes": release_notes,
                            "online": target_online,
                            "error": target_error,
                            "disabled": target_disabled,
                            "compatible": target_compatible,
                            "releaseChannels": {},
                        }

                        if (
                            populated_check["type"] == "github_release"
                            and "stable_branch" in populated_check
                            and "prerelease_branches" in populated_check
                        ):
                            # target supports release channels via github branches and releases
                            def to_release_channel(branch_info):
                                return {
                                    "name": branch_info["name"],
                                    "channel": branch_info["branch"],
                                }

                            stable_channel = populated_check["stable_branch"]["branch"]
                            current_channel = populated_check.get("prerelease_channel")
                            release_channels = [
                                to_release_channel(populated_check["stable_branch"]),
                            ] + list(
                                map(
                                    to_release_channel,
                                    populated_check["prerelease_branches"],
                                )
                            )
                            information[target]["releaseChannels"] = {
                                "available": release_channels,
                                "current": current_channel
                                if current_channel
                                else stable_channel,
                                "stable": stable_channel,
                            }

                        if "released_version" in populated_check:
                            information[target]["releasedVersion"] = populated_check[
                                "released_version"
                            ]

                if self._version_cache_dirty:
                    self._save_version_cache()

                self._get_versions_data = information, update_available, update_possible
                self._get_versions_data_ready.set()
            finally:
                self._get_versions_mutex.release()

        else:  # something's already in progress, let's wait for it to complete and use its result
            self._get_versions_data_ready.wait()
            information, update_available, update_possible = self._get_versions_data

        return information, update_available, update_possible

    def _get_release_notes(self, information, populated_check):
        release_notes = None

        if information and information["remote"] and information["remote"]["value"]:
            if populated_check.get("release_notes"):
                release_notes = populated_check["release_notes"]
            elif "release_notes" in information["remote"]:
                release_notes = information["remote"]["release_notes"]

            if release_notes:
                release_notes = release_notes.format(
                    octoprint_version=VERSION,
                    target_name=information["remote"]["name"],
                    target_version=information["remote"]["value"],
                )

        return release_notes

    def _get_check_hash(self, check):
        def dict_to_sorted_repr(d):
            lines = []
            for key in sorted(d.keys()):
                value = d[key]
                if isinstance(value, dict):
                    lines.append(f"{key!r}: {dict_to_sorted_repr(value)}")
                else:
                    lines.append(f"{key!r}: {value!r}")

            return "{" + ", ".join(lines) + "}"

        hash = hashlib.md5()

        def hash_update(value):
            value = value.encode("utf-8")
            hash.update(value)

        hash_update(dict_to_sorted_repr(check))
        return hash.hexdigest()

    def _get_current_version(
        self, target, check, force=False, online=None, credentials=None
    ):
        """
        Determines the current version information for one target based on its check configuration.
        """

        current_hash = self._get_check_hash(check)
        if online is None:
            online = self._connectivity_checker.online
        if target in self._version_cache and not force:
            data = self._version_cache[target]
            if (
                data["hash"] == current_hash
                and data["timestamp"] + self._version_cache_ttl
                >= time.time()
                > data["timestamp"]
                and data.get("online", None) == online
            ):
                # we also check that timestamp < now to not get confused too much by clock changes
                return (
                    data["information"],
                    data["available"],
                    data["possible"],
                    data["online"],
                    data.get("error", None),
                )

        information = {}
        update_available = False
        error = None

        disabled = check.get("disabled", False)
        compatible = check.get("compatible", True)

        try:
            version_checker = self._get_version_checker(target, check)
            information, is_current = version_checker.get_latest(
                target, check, online=online, credentials=credentials
            )
            if information is not None:
                if (
                    is_current
                    and check["type"] in self.CURRENT_TRACKING_TYPES
                    and check["current"] is None
                ):
                    self._persist_check_current(
                        target, check, information["remote"]["value"]
                    )
                    del self._version_cache[target]
                    self._version_cache_dirty = True
                elif not is_current and compatible:
                    update_available = True
        except exceptions.CannotCheckOffline:
            update_possible = False
            information["needs_online"] = True
        except exceptions.UnknownCheckType:
            self._logger.warning(
                "Unknown check type {} for {}".format(check["type"], target)
            )
            update_possible = False
            error = "unknown_check"
        except exceptions.NetworkError:
            self._logger.warning(
                f"Could not check {target} for updates due to a network error"
            )
            update_possible = False
            error = "network"
        except exceptions.ApiCheckError as exc:
            self._logger.warning(
                f"Could not check {target} for updates due to an API error: {exc}"
            )
            update_possible = False
            error = "api"
        except exceptions.RateLimitCheckError as exc:
            self._logger.warning(
                "Could not check {} for updates due to running into a rate limit: {}".format(
                    target, exc
                )
            )
            update_possible = False
            error = "ratelimit"
        except exceptions.CheckError:
            self._logger.warning(
                f"Could not check {target} for updates due to a check error"
            )
            update_possible = False
            error = "check"
        except Exception:
            self._logger.exception(f"Could not check {target} for updates")
            update_possible = False
            error = "unknown"
        else:
            try:
                updater = self._get_updater(target, check)
                update_possible = compatible and updater.can_perform_update(
                    target, check, online=online
                )
            except Exception:
                self._logger.exception(f"Error while checking if {target} can be updated")
                update_possible = False

        self._version_cache[target] = {
            "timestamp": time.time(),
            "hash": current_hash,
            "information": information,
            "available": update_available,
            "possible": update_possible,
            "online": online,
            "error": error,
            "disabled": disabled,
            "compatible": compatible,
        }
        self._version_cache_dirty = True
        return information, update_available, update_possible, online, error

    def _queued_updates_timer_start(self):
        if self._queued_updates_abort_timer is not None:
            return

        self._logger.debug("Starting queued updates timer.")

        self._timeout_value = 60
        self._queued_updates_abort_timer = RepeatedTimer(
            1, self._queued_updates_timer_task
        )
        self._queued_updates_abort_timer.start()

    def _queued_updates_timer_stop(self):
        if self._queued_updates_abort_timer is not None:
            self._queued_updates_abort_timer.cancel()
            self._queued_updates_abort_timer = None
            self._send_client_message(
                "queued_updates",
                {
                    "targets": self._queued_updates["targets"],
                    "timeout_value": -1,
                },
            )

    def _queued_updates_timer_task(self):
        if self._timeout_value is None:
            return

        self._timeout_value -= 1
        if self._timeout_value <= 0:
            if self._queued_updates_abort_timer is not None:
                self._queued_updates_abort_timer.cancel()
                self._queued_updates_abort_timer = None
                self.perform_updates(
                    targets=self._queued_updates["targets"],
                    force=self._queued_updates["force"],
                )
                self._queued_updates = {"targets": [], "force": True}
        self._send_client_message(
            "queued_updates",
            {
                "targets": self._queued_updates["targets"],
                "timeout_value": self._timeout_value,
            },
        )

    def perform_updates(self, force=False, **kwargs):
        """
        Performs the updates for the given check_targets. Will update all possible targets by default.

        :param targets: an iterable defining the targets to update, if not supplied defaults to all targets
        """

        if not self._environment_supported:
            self._logger.error("Direct updates are unsupported in this environment")
            return [], {}

        if not self._storage_sufficient:
            self._logger.error("Not enough free disk space for updating")
            return [], {}

        targets = kwargs.get("targets", kwargs.get("check_targets", None))

        checks = self._get_configured_checks()
        populated_checks = {}
        for target, check in checks.items():
            try:
                populated_checks[target] = self._populated_check(target, check)
            except exceptions.UnknownCheckType:
                self._logger.debug(f"Ignoring unknown check type for target {target}")
            except Exception:
                self._logger.exception(
                    "Error while populating check prior to update for target {}".format(
                        target
                    )
                )

        if targets is None:
            targets = populated_checks.keys()
        to_be_updated = sorted(set(targets) & set(populated_checks.keys()))
        if "octoprint" in to_be_updated:
            to_be_updated.remove("octoprint")
            tmp = ["octoprint"] + to_be_updated
            to_be_updated = tmp
        if "pip" in to_be_updated:
            to_be_updated.remove("pip")
            tmp = ["pip"] + to_be_updated
            to_be_updated = tmp

        updater_thread = threading.Thread(
            target=self._update_worker, args=(populated_checks, to_be_updated, force)
        )
        updater_thread.daemon = False
        updater_thread.start()

        check_data = {
            key: check["displayName"] if "displayName" in check else key
            for key, check in populated_checks.items()
            if key in to_be_updated
        }
        return to_be_updated, check_data

    def _update_worker(self, checks, check_targets, force):

        restart_type = None

        credentials = self._settings.get(["credentials"], merged=True)

        try:
            self._update_in_progress = True

            target_results = {}
            error = False

            ### iterate over all configured targets

            for target in check_targets:
                if target not in checks:
                    continue
                check = checks[target]

                if not check.get("enabled", True) or not check.get("compatible", True):
                    continue

                if target not in check_targets:
                    continue

                target_error, target_result = self._perform_update(
                    target, check, force, credentials=credentials
                )
                error = error or target_error
                if target_result is not None:
                    target_results[target] = target_result

                    target_restart_type = self._get_restart_type(check)

                    # if our update requires a restart we have to determine which type
                    if restart_type is None or (
                        restart_type == "octoprint"
                        and target_restart_type == "environment"
                    ):
                        restart_type = target_restart_type

        finally:
            # we might have needed to update the config, so we'll save that now
            self._settings.save()

            # we also need to save our update log changes
            self._save_update_log()

            # also, we are now longer updating
            self._update_in_progress = False

        if error:
            # if there was an unignorable error, we just return error
            self._send_client_message("error", {"results": target_results})

        else:
            self._save_version_cache()

            # otherwise the update process was a success, but we might still have to restart
            if restart_type is not None and restart_type in ("octoprint", "environment"):
                # one of our updates requires a restart of either type "octoprint" or "environment". Let's see if
                # we can actually perform that

                if self._has_restart_command(restart_type):
                    self._send_client_message(
                        "restarting",
                        {"restart_type": restart_type, "results": target_results},
                    )
                    try:
                        self._perform_restart(restart_type)
                    except exceptions.RestartFailed:
                        self._send_client_message(
                            "restart_failed",
                            {"restart_type": restart_type, "results": target_results},
                        )
                else:
                    # we don't have this restart type configured, we'll have to display a message that a manual
                    # restart is needed
                    self._send_client_message(
                        "restart_manually",
                        {"restart_type": restart_type, "results": target_results},
                    )
            else:
                self._send_client_message("success", {"results": target_results})

    def _perform_update(self, target, check, force, credentials=None):
        online = self._connectivity_checker.online

        information, update_available, update_possible, _, _ = self._get_current_version(
            target, check, online=online, credentials=credentials
        )
        populated_check = self._populated_check(target, check)

        def add_update_log(success, text):
            self._add_update_log_entry(
                target,
                information["local"]["value"],
                information["remote"]["value"],
                success,
                text,
                display_name=populated_check["displayName"],
                release_notes=self._get_release_notes(information, populated_check),
                save=False,
            )

        if not update_available and not force:
            add_update_log(False, "Target is already up to date")
            return False, None

        if not update_possible:
            self._logger.warning(
                "Cannot perform update for {}, update type is not fully configured".format(
                    target
                )
            )
            add_update_log(False, "Update not possible, missing information")
            return False, None

        if not information.get("compatible", True):
            self._logger.warning(
                "Cannot perform update for {}, update is reported as incompatible".format(
                    target
                )
            )
            add_update_log(False, "Update is incompatible")
            return False, None

        # determine the target version to update to
        target_version = information["remote"]["value"]
        target_error = False
        target_result = None

        def trigger_event(success, **additional_payload):
            from octoprint.events import Events

            if success:
                # noinspection PyUnresolvedReferences
                event = Events.PLUGIN_SOFTWAREUPDATE_UPDATE_SUCCEEDED
            else:
                # noinspection PyUnresolvedReferences
                event = Events.PLUGIN_SOFTWAREUPDATE_UPDATE_FAILED

            payload = copy.copy(additional_payload)
            payload.update(
                {
                    "target": target,
                    "from_version": information["local"]["value"],
                    "to_version": target_version,
                }
            )

            self._event_bus.fire(event, payload=payload)

        ### The actual update procedure starts here...

        try:
            self._logger.info(f"Starting update of {target} to {target_version}...")
            self._send_client_message(
                "updating",
                {
                    "target": target,
                    "version": target_version,
                    "name": populated_check["displayName"],
                },
            )
            updater = self._get_updater(target, check)
            if updater is None:
                raise exceptions.UnknownUpdateType()

            update_result = updater.perform_update(
                target,
                populated_check,
                target_version,
                log_cb=self._log,
                online=online,
                force=force,
            )
            target_result = ("success", update_result)
            self._logger.info(f"Update of {target} to {target_version} successful!")
            trigger_event(True)

        except exceptions.UnknownUpdateType:
            self._logger.warning(
                "Update of %s can not be performed, unknown update type" % target
            )
            self._send_client_message(
                "update_failed",
                {
                    "target": target,
                    "version": target_version,
                    "name": populated_check["displayName"],
                    "reason": "Unknown update type",
                },
            )
            add_update_log(False, "Update failed, unknown update type")
            return False, None

        except exceptions.CannotUpdateOffline:
            self._logger.warning(
                "Update of %s can not be performed, it's not marked as 'offline' capable but we are apparently offline right now"
                % target
            )
            self._send_client_message(
                "update_failed",
                {
                    "target": target,
                    "version": target_version,
                    "name": populated_check["displayName"],
                    "reason": "No internet connection",
                },
            )
            add_update_log(False, "No internet connection")
            return False, None

        except Exception as e:
            self._logger.exception(
                "Update of %s can not be performed, please also check plugin_softwareupdate_console.log for possible causes of this"
                % target
            )
            trigger_event(False)

            if "ignorable" not in populated_check or not populated_check["ignorable"]:
                target_error = True

            if isinstance(e, exceptions.UpdateError):
                target_result = ("failed", e.data)
                self._send_client_message(
                    "update_failed",
                    {
                        "target": target,
                        "version": target_version,
                        "name": populated_check["displayName"],
                        "reason": e.data,
                    },
                )
            else:
                target_result = ("failed", None)
                self._send_client_message(
                    "update_failed",
                    {
                        "target": target,
                        "version": target_version,
                        "name": populated_check["displayName"],
                        "reason": "unknown",
                    },
                )
            add_update_log(False, "Update failed")

        else:
            # make sure that any external changes to config.yaml are loaded into the system
            self._settings.load()

            # persist the new version if necessary for check type
            self._persist_check_current(target, check, target_version)

            del self._version_cache[target]
            self._version_cache_dirty = True

            add_update_log(True, "Update successful")

        return target_error, target_result

    def _persist_check_current(self, target, check, current):
        if check["type"] not in self.CURRENT_TRACKING_TYPES:
            return

        self._settings.load()

        dummy_default = {"plugins": {}}
        dummy_default["plugins"][self._identifier] = {"checks": {}}
        dummy_default["plugins"][self._identifier]["checks"][target] = {"current": None}
        self._settings.set(["checks", target, "current"], current, defaults=dummy_default)

        # we have to save here (even though that makes us save quite often) since otherwise the next
        # load will overwrite our changes we just made
        self._settings.save()

        check["current"] = current

    def _perform_restart(self, restart_type):
        """
        Performs a restart using the supplied restart_type.
        """

        self._logger.info("Restarting...")
        try:
            if restart_type == "octoprint":
                return self._system_commands.perform_server_restart()
            elif restart_type == "environment":
                return self._system_commands.perform_system_restart()
        except CommandlineError as e:
            self._logger.exception(f"Error while restarting of type {restart_type}")
            self._logger.warning(f"Restart stdout:\n{e.stdout}")
            self._logger.warning(f"Restart stderr:\n{e.stderr}")
            raise exceptions.RestartFailed()

    def _populated_check(self, target, check):
        from flask_babel import gettext

        if "type" not in check:
            raise exceptions.UnknownCheckType()

        result = dict(check)

        if target == "octoprint":
            displayName = check.get("displayName")
            if displayName is None:
                # displayName missing or set to None
                displayName = gettext("OctoPrint")
            result["displayName"] = to_unicode(displayName, errors="replace")

            displayVersion = check.get("displayVersion")
            if displayVersion is None:
                # displayVersion missing or set to None
                displayVersion = "{octoprint_version}"
            result["displayVersion"] = to_unicode(displayVersion, errors="replace")

            result["released_version"] = is_released_octoprint_version()

            if check["type"] in self.COMMIT_TRACKING_TYPES:
                result["current"] = REVISION if REVISION else "unknown"
            else:
                result["current"] = VERSION

        elif target == "pip":
            import pkg_resources

            displayName = check.get("displayName")
            if displayName is None:
                # displayName missing or set to None
                displayName = gettext("pip")
            result["displayName"] = to_unicode(displayName, errors="replace")

            displayVersion = check.get("displayVersion")
            if displayVersion is None:
                # displayVersion missing or set to None
                distribution = pkg_resources.get_distribution("pip")
                if distribution:
                    displayVersion = distribution.version
            result["displayVersion"] = to_unicode(displayVersion, errors="replace")

            result["pip_command"] = check.get(
                "pip_command",
                self._settings.global_get(["server", "commands", "localPipCommand"]),
            )

        else:
            result["displayName"] = to_unicode(check.get("displayName"), errors="replace")
            if result["displayName"] is None:
                # displayName missing or None
                result["displayName"] = to_unicode(target, errors="replace")

            result["displayVersion"] = to_unicode(
                check.get("displayVersion", check.get("current")), errors="replace"
            )
            if result["displayVersion"] is None:
                # displayVersion AND current missing or None
                result["displayVersion"] = "unknown"

            if check["type"] in self.CURRENT_TRACKING_TYPES:
                result["current"] = check.get("current", None)
            else:
                result["current"] = check.get(
                    "current", check.get("displayVersion", None)
                )

        if (
            check["type"] in self.RELEASE_TRACKING_TYPES
            and result["current"]
            and (check.get("prerelease", None) or not is_stable(result["current"]))
        ):
            # we are tracking releases and are either also tracking prerelease OR are currently running
            # a non stable version => we need to change some parameters

            # we compare versions fully, not just the base so that we see a difference
            # between RCs + stable for the same version release
            result["force_base"] = False

            if check["type"] == "github_release":
                if check.get("prerelease", None):
                    # we are tracking prereleases => we want to be on the correct prerelease channel/branch
                    channel = check.get("prerelease_channel", None)
                    if channel:
                        # if we have a release channel, we also set our update_branch here to our release channel
                        # in case it's not already set
                        result["update_branch"] = check.get("update_branch", channel)

                else:
                    # we are not tracking prereleases, but aren't on the stable branch either => switch back
                    # to stable branch on update
                    result["update_branch"] = check.get(
                        "update_branch",
                        check.get("stable_branch", {"branch": "main"})["branch"],
                    )

            if check.get("update_script", None):
                # we force an exact version & python unequality check, to be able to downgrade
                result["force_exact_version"] = True
                result["release_compare"] = "python_unequal"
            elif check.get("pip", None):
                # we force python unequality check for pip installs, to be able to downgrade
                result["release_compare"] = "python_unequal"

        if result.get("pip", None):
            if "pip_command" not in result:
                local_pip_command = self._settings.global_get(
                    ["server", "commands", "localPipCommand"]
                )
                if local_pip_command:
                    result["pip_command"] = local_pip_command

        return result

    def _log(self, lines, prefix=None, stream=None, strip=True):
        if strip:
            lines = list(map(lambda x: x.strip(), lines))

        self._send_client_message(
            "loglines",
            data={"loglines": [{"line": line, "stream": stream} for line in lines]},
        )
        for line in lines:
            self._console_logger.debug(f"{prefix} {line}")

    def _send_client_message(self, message_type, data=None):
        self._plugin_manager.send_plugin_message(
            self._identifier, {"type": message_type, "data": data}
        )

    def _get_version_checker(self, target, check):
        """
        Retrieves the version checker to use for given target and check configuration. Will raise an UnknownCheckType
        if version checker cannot be determined.
        """

        if "type" not in check:
            raise exceptions.ConfigurationInvalid("no check type defined")

        check_type = check["type"]
        method = getattr(version_checks, check_type)
        if method is None:
            raise exceptions.UnknownCheckType()
        else:
            return method

    def _get_update_method(self, target, check, valid_methods=None):
        """
        Determines the update method for the given target and check.

        If ``valid_methods`` is provided, determine method must be contained
        therein to be considered valid.

        Raises an ``UnknownUpdateType`` exception if method cannot be determined
        or validated.
        """

        method = None
        if "method" in check:
            method = check["method"]
        else:
            if "update_script" in check:
                method = "update_script"
            elif "pip" in check:
                method = "pip"
            elif "python_updater" in check:
                method = "python_updater"

        if method is None or (valid_methods and method not in valid_methods):
            raise exceptions.UnknownUpdateType()

        return method

    def _get_updater(self, target, check):
        """
        Retrieves the updater for the given target and check configuration. Will raise an UnknownUpdateType if updater
        cannot be determined.
        """

        method = self._get_update_method(target, check)
        if method is None:
            raise exceptions.UnknownUpdateType()

        updater = getattr(updaters, method)
        if updater is None:
            raise exceptions.UnknownUpdateType()

        return updater

    def _get_octoprint_checkout_folder(self, checks=None):
        if checks is None:
            checks = self._get_configured_checks()

        if "octoprint" not in checks:
            return None

        if "checkout_folder" in checks["octoprint"]:
            return checks["octoprint"]["checkout_folder"]
        elif "update_folder" in checks["octoprint"]:
            return checks["octoprint"]["update_folder"]

        return None

    def _get_octoprint_tracked_branch(self, checks=None):
        if checks is None:
            checks = self._get_configured_checks()

        if "octoprint" not in checks:
            return None

        return checks["octoprint"].get("branch")

    def _get_octoprint_pip_target(self, checks=None):
        if checks is None:
            checks = self._get_configured_checks()

        if "octoprint" not in checks:
            return None

        return checks["octoprint"].get("pip")


def _register_custom_events(*args, **kwargs):
    return ["update_succeeded", "update_failed"]


__plugin_name__ = "Software Update"
__plugin_author__ = "Gina Häußge"
__plugin_url__ = "https://docs.octoprint.org/en/master/bundledplugins/softwareupdate.html"
__plugin_description__ = "Allows receiving update notifications and performing updates of OctoPrint and plugins"
__plugin_disabling_discouraged__ = gettext(
    "Without this plugin OctoPrint will no longer be able to "
    "update itself or any of your installed plugins which might put "
    "your system at risk."
)
__plugin_license__ = "AGPLv3"
__plugin_pythoncompat__ = ">=3.7,<4"


def __plugin_load__():
    global __plugin_implementation__
    __plugin_implementation__ = SoftwareUpdatePlugin()

    global __plugin_helpers__
    __plugin_helpers__ = {
        "version_checks": version_checks,
        "updaters": updaters,
        "exceptions": exceptions,
        "util": util,
    }

    global __plugin_hooks__
    __plugin_hooks__ = {
        "octoprint.cli.commands": cli.commands,
        "octoprint.events.register_custom_events": _register_custom_events,
        "octoprint.access.permissions": __plugin_implementation__.get_additional_permissions,
        "octoprint.systeminfo.additional_bundle_files": __plugin_implementation__.get_additional_bundle_files,
    }
